#!/usr/bin/env python3

from typing import Optional, Union

import torch
from linear_operator.operators import KroneckerProductLinearOperator, RootLinearOperator
from linear_operator.utils.interpolation import left_interp
from torch import LongTensor, Tensor

from .. import settings
from ..distributions import MultitaskMultivariateNormal, MultivariateNormal
from ..module import Module
from ._variational_strategy import _VariationalStrategy


def _select_lmc_coefficients(lmc_coefficients: torch.Tensor, indices: torch.LongTensor) -> torch.Tensor:
    """
    Given a list of indices for ... x N datapoints,
      select the row from lmc_coefficient that corresponds to each datapoint

    lmc_coefficients: torch.Tensor ... x num_latents x ... x num_tasks
    indices: torch.Tesnor ... x N
    """
    batch_shape = torch.broadcast_shapes(lmc_coefficients.shape[:-1], indices.shape[:-1])

    # We will use the left_interp helper to do the indexing
    lmc_coefficients = lmc_coefficients.expand(*batch_shape, lmc_coefficients.shape[-1])[..., None]
    indices = indices.expand(*batch_shape, indices.shape[-1])[..., None]
    res = left_interp(
        indices,
        torch.ones(indices.shape, dtype=torch.long, device=indices.device),
        lmc_coefficients,
    ).squeeze(-1)
    return res


class LMCVariationalStrategy(_VariationalStrategy):
    r"""
    LMCVariationalStrategy is an implementation of the "Linear Model of Coregionalization"
    for multitask GPs. This model assumes that there are :math:`Q` latent functions
    :math:`\mathbf g(\cdot) = [g^{(1)}(\cdot), \ldots, g^{(q)}(\cdot)]`,
    each of which is modelled by a GP.
    The output functions (tasks) are linear combination of the latent functions:

    .. math::

        f_{\text{task } i}( \mathbf x) = \sum_{q=1}^Q a_i^{(q)} g^{(q)} ( \mathbf x )

    LMCVariationalStrategy wraps an existing :obj:`~gpytorch.variational.VariationalStrategy`.
    The output will either be a :obj:`~gpytorch.distributions.MultitaskMultivariateNormal` distribution
    (if we wish to evaluate all tasks for each input) or a :obj:`~gpytorch.distributions.MultivariateNormal`
    (if we wish to evaluate a single task for each input).

    The base variational strategy is assumed to operate on a multi-batch of GPs, where one
    of the batch dimensions corresponds to the latent function dimension.

    .. note::

        The batch shape of the base :obj:`~gpytorch.variational.VariationalStrategy` does not
        necessarily have to correspond to the batch shape of the underlying GP objects.

        For example, if the base variational strategy has a batch shape of `[3]` (corresponding
        to 3 latent functions), the GP kernel object could have a batch shape of `[3]` or no
        batch shape. This would correspond to each of the latent functions having different kernels
        or the same kernel, respectivly.

    Example:
        >>> class LMCMultitaskGP(gpytorch.models.ApproximateGP):
        >>>     '''
        >>>     3 latent functions
        >>>     5 output dimensions (tasks)
        >>>     '''
        >>>     def __init__(self):
        >>>         # Each latent function shares the same inducing points
        >>>         # We'll have 32 inducing points, and let's assume the input dimensionality is 2
        >>>         inducing_points = torch.randn(32, 2)
        >>>
        >>>         # The variational parameters have a batch_shape of [3] - for 3 latent functions
        >>>         variational_distribution = gpytorch.variational.MeanFieldVariationalDistribution(
        >>>             inducing_points.size(-1), batch_shape=torch.Size([3]),
        >>>         )
        >>>         variational_strategy = gpytorch.variational.LMCVariationalStrategy(
        >>>             gpytorch.variational.VariationalStrategy(
        >>>                 self, inducing_points, variational_distribution, learn_inducing_locations=True,
        >>>             ),
        >>>             num_tasks=5,
        >>>             num_latents=3,
        >>>             latent_dim=-1,
        >>>         )
        >>>
        >>>         # Each latent function has its own mean/kernel function
        >>>         super().__init__(variational_strategy)
        >>>         self.mean_module = gpytorch.means.ConstantMean(batch_shape=torch.Size([3]))
        >>>         self.covar_module = gpytorch.kernels.ScaleKernel(
        >>>             gpytorch.kernels.RBFKernel(batch_shape=torch.Size([3])),
        >>>             batch_shape=torch.Size([3]),
        >>>         )
        >>>

    :param base_variational_strategy: Base variational strategy
    :param num_tasks: The total number of tasks (output functions)
    :param num_latents: The total number of latent functions in each group
    :param latent_dim: (Default: -1) Which batch dimension corresponds to the latent function batch.
        **Must be negative indexed**
    :param jitter_val: Amount of diagonal jitter to add for Cholesky factorization numerical stability
    """

    def __init__(
        self,
        base_variational_strategy: _VariationalStrategy,
        num_tasks: int,
        num_latents: int = 1,
        latent_dim: int = -1,
        jitter_val: Optional[float] = None,
    ):
        Module.__init__(self)
        self.base_variational_strategy = base_variational_strategy
        self.num_tasks = num_tasks
        batch_shape = self.base_variational_strategy._variational_distribution.batch_shape

        # Check if no functions
        if latent_dim >= 0:
            raise RuntimeError(f"latent_dim must be a negative indexed batch dimension: got {latent_dim}.")
        if not (batch_shape[latent_dim] == num_latents or batch_shape[latent_dim] == 1):
            raise RuntimeError(
                f"Mismatch in num_latents: got a variational distribution of batch shape {batch_shape}, "
                f"expected the function dim {latent_dim} to be {num_latents}."
            )
        self.num_latents = num_latents
        self.latent_dim = latent_dim

        # Make the batch_shape
        self.batch_shape = list(batch_shape)
        del self.batch_shape[self.latent_dim]
        self.batch_shape = torch.Size(self.batch_shape)

        # LCM coefficients
        lmc_coefficients = torch.randn(*batch_shape, self.num_tasks)
        self.register_parameter("lmc_coefficients", torch.nn.Parameter(lmc_coefficients))

        if jitter_val is None:
            self.jitter_val = settings.variational_cholesky_jitter.value(
                self.base_variational_strategy.inducing_points.dtype
            )
        else:
            self.jitter_val = jitter_val

    @property
    def prior_distribution(self) -> MultivariateNormal:
        return self.base_variational_strategy.prior_distribution

    @property
    def variational_distribution(self) -> MultivariateNormal:
        return self.base_variational_strategy.variational_distribution

    @property
    def variational_params_initialized(self) -> bool:
        return self.base_variational_strategy.variational_params_initialized

    def kl_divergence(self) -> Tensor:
        return super().kl_divergence().sum(dim=self.latent_dim)

    def __call__(
        self, x: Tensor, prior: bool = False, task_indices: Optional[LongTensor] = None, **kwargs
    ) -> Union[MultitaskMultivariateNormal, MultivariateNormal]:
        r"""
        Computes the variational (or prior) distribution
        :math:`q( \mathbf f \mid \mathbf X)` (or :math:`p( \mathbf f \mid \mathbf X)`).
        There are two modes:

        1.  Compute **all tasks** for all inputs.
            If this is the case, the task_indices attribute should be None.
            The return type will be a (... x N x num_tasks)
            :class:`~gpytorch.distributions.MultitaskMultivariateNormal`.
        2.  Compute **one task** per inputs.
            If this is the case, the (... x N) task_indices tensor should contain
            the indices of each input's assigned task.
            The return type will be a (... x N)
            :class:`~gpytorch.distributions.MultivariateNormal`.

        :param x: (... x N x D) Input locations to evaluate variational strategy
        :param task_indices: (Default: None) Task index associated with each input.
            If this **is not** provided, then the returned distribution evaluates every input on every task
            (returns :class:`~gpytorch.distributions.MultitaskMultivariateNormal`).
            If this **is** provided, then the returned distribution evaluates each input only on its assigned task.
            (returns :class:`~gpytorch.distributions.MultivariateNormal`).
        :param prior: (Default: False) If False, returns the variational distribution
            :math:`q( \mathbf f \mid \mathbf X)`.
            If True, returns the prior distribution
            :math:`p( \mathbf f \mid \mathbf X)`.
        :return: :math:`q( \mathbf f \mid \mathbf X)` (or the prior),
            either for all tasks (if `task_indices == None`)
            or for a specific task (if `task_indices != None`).
        :rtype: ~gpytorch.distributions.MultitaskMultivariateNormal (... x N x num_tasks)
            or ~gpytorch.distributions.MultivariateNormal (... x N)
        """
        latent_dist = self.base_variational_strategy(x, prior=prior, **kwargs)
        num_batch = len(latent_dist.batch_shape)
        latent_dim = num_batch + self.latent_dim

        if task_indices is None:
            num_dim = num_batch + len(latent_dist.event_shape)

            # Every data point will get an output for each task
            # Therefore, we will set up the lmc_coefficients shape for a matmul
            lmc_coefficients = self.lmc_coefficients.expand(*latent_dist.batch_shape, self.lmc_coefficients.size(-1))

            # Mean: ... x N x num_tasks
            latent_mean = latent_dist.mean.permute(*range(0, latent_dim), *range(latent_dim + 1, num_dim), latent_dim)
            mean = latent_mean @ lmc_coefficients.permute(
                *range(0, latent_dim), *range(latent_dim + 1, num_dim - 1), latent_dim, -1
            )

            # Covar: ... x (N x num_tasks) x (N x num_tasks)
            latent_covar = latent_dist.lazy_covariance_matrix
            lmc_factor = RootLinearOperator(lmc_coefficients.unsqueeze(-1))
            covar = KroneckerProductLinearOperator(latent_covar, lmc_factor).sum(latent_dim)
            # Add a bit of jitter to make the covar PD
            covar = covar.add_jitter(self.jitter_val)

            # Done!
            function_dist = MultitaskMultivariateNormal(mean, covar)

        else:
            # Each data point will get a single output corresponding to a single task
            # Therefore, we will select the appropriate lmc coefficients for each task
            lmc_coefficients = _select_lmc_coefficients(self.lmc_coefficients, task_indices)

            # Mean: ... x N
            mean = (latent_dist.mean * lmc_coefficients).sum(latent_dim)

            # Covar: ... x N x N
            latent_covar = latent_dist.lazy_covariance_matrix
            lmc_factor = RootLinearOperator(lmc_coefficients.unsqueeze(-1))
            covar = (latent_covar * lmc_factor).sum(latent_dim)
            # Add a bit of jitter to make the covar PD
            covar = covar.add_jitter(self.jitter_val)

            # Done!
            function_dist = MultivariateNormal(mean, covar)

        return function_dist



#!/usr/bin/env python3

import unittest

import torch

import gpytorch
from gpytorch import settings
from gpytorch.kernels import GridInterpolationKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood
from gpytorch.models import ExactGP
from gpytorch.models.exact_prediction_strategies import InterpolatedPredictionStrategy
from gpytorch.test.model_test_case import BaseModelTestCase

N_PTS = 50


class GridInterpolationKernelMock(GridInterpolationKernel):
    def __init__(self, should_use_wiski=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.should_use_wiski = should_use_wiski

    def prediction_strategy(self, *args, **kwargs):
        return InterpolatedPredictionStrategy(uses_wiski=self.should_use_wiski, *args, **kwargs)


class ExactGPModel(ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class InterpolatedExactGPModel(ExactGP):
    def __init__(self, train_x, train_y, likelihood, should_use_wiski=False):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = GridInterpolationKernelMock(
            base_kernel=gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel()),
            grid_size=128,
            num_dims=1,
            should_use_wiski=should_use_wiski,
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class SumExactGPModel(ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        covar_a = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        covar_b = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel(nu=0.5))
        covar_c = gpytorch.kernels.LinearKernel()  # this one is important because its covariance matrix can be lazy
        self.covar_module = covar_a + covar_b + covar_c

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class TestExactGP(BaseModelTestCase, unittest.TestCase):
    def create_model(self, train_x, train_y, likelihood):
        model = ExactGPModel(train_x, train_y, likelihood)
        return model

    def create_test_data(self):
        return torch.randn(N_PTS, 1)

    def create_likelihood_and_labels(self):
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        labels = torch.randn(N_PTS) + 2
        return likelihood, labels

    def create_batch_test_data(self, batch_shape=torch.Size([3])):
        return torch.randn(*batch_shape, N_PTS, 1)

    def create_batch_likelihood_and_labels(self, batch_shape=torch.Size([3])):
        likelihood = gpytorch.likelihoods.GaussianLikelihood(batch_shape=batch_shape)
        labels = torch.randn(*batch_shape, N_PTS) + 2
        return likelihood, labels

    def test_forward_eval_fast(self):
        with gpytorch.settings.max_eager_kernel_size(1), gpytorch.settings.fast_pred_var(True):
            self.test_forward_eval()

    def test_batch_forward_eval_fast(self):
        with gpytorch.settings.max_eager_kernel_size(1), gpytorch.settings.fast_pred_var(True):
            self.test_batch_forward_eval()

    def test_multi_batch_forward_eval_fast(self):
        with gpytorch.settings.max_eager_kernel_size(1), gpytorch.settings.fast_pred_var(True):
            self.test_multi_batch_forward_eval()

    def test_batch_forward_then_nonbatch_forward_eval(self):
        batch_data = self.create_batch_test_data()
        likelihood, labels = self.create_batch_likelihood_and_labels()
        model = self.create_model(batch_data, labels, likelihood)
        model.eval()
        output = model(batch_data)

        # Smoke test derivatives working
        output.mean.sum().backward()

        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == batch_data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == batch_data.size(-2))

        # Create non-batch data
        data = self.create_test_data()
        output = model(data)
        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == data.size(-2))

        # Smoke test derivatives working
        output.mean.sum().backward()

    def test_batch_forward_then_different_batch_forward_eval(self):
        non_batch_data = self.create_test_data()
        likelihood, labels = self.create_likelihood_and_labels()
        model = self.create_model(non_batch_data, labels, likelihood)
        model.eval()

        # Batch size 3
        batch_data = self.create_batch_test_data()
        output = model(batch_data)
        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == batch_data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == batch_data.size(-2))

        # Now Batch size 2
        batch_data = self.create_batch_test_data(batch_shape=torch.Size([2]))
        output = model(batch_data)
        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == batch_data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == batch_data.size(-2))

        # Now 3 again
        batch_data = self.create_batch_test_data()
        output = model(batch_data)
        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == batch_data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == batch_data.size(-2))

        # Now 1
        batch_data = self.create_batch_test_data(batch_shape=torch.Size([1]))
        output = model(batch_data)
        self.assertTrue(output.lazy_covariance_matrix.dim() == 3)
        self.assertTrue(output.lazy_covariance_matrix.size(-1) == batch_data.size(-2))
        self.assertTrue(output.lazy_covariance_matrix.size(-2) == batch_data.size(-2))

    def test_prior_mode(self):
        train_data = self.create_test_data()
        likelihood, labels = self.create_likelihood_and_labels()
        prior_model = self.create_model(None, None, likelihood)
        model = self.create_model(train_data, labels, likelihood)
        prior_model.eval()
        model.eval()

        test_data = self.create_test_data()
        prior_out = prior_model(test_data)
        with gpytorch.settings.prior_mode(True):
            prior_out_cm = model(test_data)
        self.assertTrue(torch.allclose(prior_out.mean, prior_out_cm.mean))
        self.assertTrue(torch.allclose(prior_out.covariance_matrix, prior_out_cm.covariance_matrix))

    def test_lanczos_fantasy_model(self):
        lanczos_thresh = 10
        n = lanczos_thresh + 1
        n_dims = 2
        with settings.max_cholesky_size(lanczos_thresh):
            x = torch.ones((n, n_dims))
            y = torch.randn(n)
            likelihood = GaussianLikelihood()
            model = ExactGPModel(x, y, likelihood=likelihood)
            mll = ExactMarginalLogLikelihood(likelihood, model)
            mll.train()
            mll.eval()

            # get a posterior to fill in caches
            model(torch.randn((1, n_dims)))

            new_n = 2
            new_x = torch.randn((new_n, n_dims))
            new_y = torch.randn(new_n)
            # just check that this can run without error
            model.get_fantasy_model(new_x, new_y)


class TestInterpolatedExactGP(TestExactGP):
    def create_model(self, train_x, train_y, likelihood):
        model = InterpolatedExactGPModel(train_x, train_y, likelihood)
        return model


class TestWiskiExactGP(TestInterpolatedExactGP):
    def create_model(self, train_x, train_y, likelihood):
        model = InterpolatedExactGPModel(train_x, train_y, likelihood, should_use_wiski=True)
        return model

    def test_fantasy_model(self):
        x = self.create_test_data()
        likelihood, labels = self.create_likelihood_and_labels()
        model = self.create_model(x, labels, likelihood)
        test_x = self.create_test_data()
        _, test_labels = self.create_likelihood_and_labels()
        with torch.no_grad():
            model.eval()
            model(test_x)
        new_model = model.get_fantasy_model(test_x, test_labels)

        self.assertEqual(type(new_model), type(model))
        self.assertTrue(new_model.prediction_strategy.uses_wiski)

    def test_nonbatch_to_batch_fantasy_model(self, batch_shape=torch.Size([3])):
        x = self.create_test_data()
        likelihood, labels = self.create_likelihood_and_labels()
        model = self.create_model(x, labels, likelihood)
        test_x = self.create_batch_test_data(batch_shape=batch_shape)
        _, test_labels = self.create_batch_likelihood_and_labels(batch_shape=batch_shape)
        with torch.no_grad():
            model.eval()
            model(test_x)
        new_model = model.get_fantasy_model(test_x, test_labels)

        self.assertEqual(type(new_model), type(model))
        self.assertTrue(new_model.prediction_strategy.uses_wiski)

    def test_nonbatch_to_multibatch_fantasy_model(self):
        self.test_nonbatch_to_batch_fantasy_model(batch_shape=torch.Size([2, 3]))


class TestSumExactGP(TestExactGP):
    def create_model(self, train_x, train_y, likelihood):
        model = SumExactGPModel(train_x, train_y, likelihood)
        return model

    def test_cache_across_lazy_threshold(self):
        x = self.create_test_data()
        likelihood, labels = self.create_likelihood_and_labels()
        model = self.create_model(x, labels, likelihood)
        model.eval()
        model(x)  # populate caches

        with settings.max_eager_kernel_size(2 * N_PTS - 1), settings.fast_pred_var(True):
            # now we'll cross the threshold and use lazy tensors
            new_x = self.create_test_data()
            _, new_y = self.create_likelihood_and_labels()
            model = model.get_fantasy_model(new_x, new_y)
            predicted = model(self.create_test_data())

            # the main purpose of the test was to ensure there was no error, but we can verify shapes too
            self.assertEqual(predicted.mean.shape, torch.Size([N_PTS]))
            self.assertEqual(predicted.variance.shape, torch.Size([N_PTS]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import math
import unittest

import torch

from gpytorch.kernels import PeriodicKernel
from gpytorch.priors import NormalPrior
from gpytorch.test.base_kernel_test_case import BaseKernelTestCase


class TestPeriodicKernel(unittest.TestCase, BaseKernelTestCase):
    def create_kernel_no_ard(self, **kwargs):
        return PeriodicKernel(**kwargs)

    def create_kernel_ard(self, num_dims, **kwargs):
        return PeriodicKernel(ard_num_dims=num_dims, **kwargs)

    def test_computes_periodic_function(self):
        a = torch.tensor([4, 2, 8], dtype=torch.float).view(3, 1)
        b = torch.tensor([0, 2], dtype=torch.float).view(2, 1)
        lengthscale = 2
        period = 3
        kernel = PeriodicKernel().initialize(lengthscale=lengthscale, period_length=period)
        kernel.eval()

        actual = torch.zeros(3, 2)
        for i in range(3):
            for j in range(2):
                val = 2 * torch.pow(torch.sin(math.pi * (a[i] - b[j]) / 3), 2) / lengthscale
                actual[i, j] = torch.exp(-val).item()

        res = kernel(a, b).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)

    def test_is_pd(self):
        # ensures 1d input is positive definite with additional jitter
        x = torch.randn(100).reshape(-1, 1)
        kernel = PeriodicKernel()
        with torch.no_grad():
            K = kernel(x, x).to_dense() + 1e-4 * torch.eye(len(x))
            eig = torch.linalg.eigvalsh(K)
            self.assertGreater(eig.min().item(), 0.0)

    def test_multidimensional_inputs(self):
        # test taken from issue #835
        # ensures multidimensional input results in a positive definite kernel matrix, with additional jitter
        x = torch.randn(1000, 2)
        kernel = PeriodicKernel()
        with torch.no_grad():
            K = kernel(x, x).to_dense() + 1e-4 * torch.eye(len(x))
            eig = torch.linalg.eigvalsh(K)
            self.assertGreaterEqual(eig.min().item(), 0.0)

    def test_batch(self):
        a = torch.tensor([[4, 2, 8], [1, 2, 3]], dtype=torch.float).view(2, 3, 1)
        b = torch.tensor([[0, 2], [-1, 2]], dtype=torch.float).view(2, 2, 1)
        period = torch.tensor(1, dtype=torch.float).view(1, 1, 1)
        lengthscale = torch.tensor(2, dtype=torch.float).view(1, 1, 1)
        kernel = PeriodicKernel().initialize(lengthscale=lengthscale, period_length=period)
        kernel.eval()

        actual = torch.zeros(2, 3, 2)
        for k in range(2):
            actual[k] = kernel(a[k], b[k]).to_dense()

        res = kernel(a, b).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)

    def test_batch_separate(self):
        a = torch.tensor([[4, 2, 8], [1, 2, 3]], dtype=torch.float).view(2, 3, 1)
        b = torch.tensor([[0, 2], [-1, 2]], dtype=torch.float).view(2, 2, 1)
        period = torch.tensor([1, 2], dtype=torch.float).view(2, 1, 1)
        lengthscale = torch.tensor([2, 1], dtype=torch.float).view(2, 1, 1)
        kernel = PeriodicKernel(batch_shape=torch.Size([2])).initialize(lengthscale=lengthscale, period_length=period)
        kernel.eval()

        actual = torch.zeros(2, 3, 2)
        for k in range(2):
            diff = a[k].unsqueeze(1) - b[k].unsqueeze(0)
            diff = diff * math.pi / period[k].unsqueeze(-1)
            actual[k] = diff.sin().pow(2).sum(dim=-1).div(lengthscale[k]).mul(-2.0).exp_()

        res = kernel(a, b).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)

    def create_kernel_with_prior(self, period_length_prior):
        return PeriodicKernel(period_length_prior=period_length_prior)

    def test_prior_type(self):
        """
        Raising TypeError if prior type is other than gpytorch.priors.Prior
        """
        kernel_fn = lambda prior: self.create_kernel_with_prior(prior)
        kernel_fn(None)
        kernel_fn(NormalPrior(0, 1))
        self.assertRaises(TypeError, kernel_fn, 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import math
import unittest
from unittest.mock import MagicMock, patch

import linear_operator
import torch
from torch import optim

import gpytorch
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.models import ApproximateGP
from gpytorch.test.base_test_case import BaseTestCase


def train_data():
    train_x = torch.linspace(0, 1, 260)
    train_y = torch.cos(train_x * (2 * math.pi))
    return train_x, train_y


class SVGPRegressionModel(ApproximateGP):
    def __init__(self, inducing_points, base_inducing_points):
        base_variational_distribution = gpytorch.variational.CholeskyVariationalDistribution(
            base_inducing_points.size(-1)
        )
        variational_distribution = gpytorch.variational.DeltaVariationalDistribution(inducing_points.size(-1))
        variational_strategy = gpytorch.variational.OrthogonallyDecoupledVariationalStrategy(
            gpytorch.variational.VariationalStrategy(
                self,
                base_inducing_points,
                base_variational_distribution,
                learn_inducing_locations=True,
                jitter_val=1e-4,
            ),
            inducing_points,
            variational_distribution,
        )
        super(SVGPRegressionModel, self).__init__(variational_strategy)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        latent_pred = gpytorch.distributions.MultivariateNormal(mean_x, covar_x)
        return latent_pred


class TestSVGPRegression(BaseTestCase, unittest.TestCase):
    seed = 0

    def test_regression_error(
        self,
        mll_cls=gpytorch.mlls.VariationalELBO,
        distribution_cls=gpytorch.variational.CholeskyVariationalDistribution,
    ):
        train_x, train_y = train_data()
        likelihood = GaussianLikelihood()
        model = SVGPRegressionModel(torch.linspace(0, 1, 128), torch.linspace(0, 1, 16))
        mll = mll_cls(likelihood, model, num_data=len(train_y))

        # Find optimal model hyperparameters
        model.train()
        likelihood.train()
        optimizer = optim.Adam([{"params": model.parameters()}, {"params": likelihood.parameters()}], lr=0.01)

        _wrapped_cg = MagicMock(wraps=linear_operator.utils.linear_cg)
        _cg_mock = patch("linear_operator.utils.linear_cg", new=_wrapped_cg)
        with _cg_mock as cg_mock:
            for _ in range(75):
                optimizer.zero_grad()
                output = model(train_x)
                loss = -mll(output, train_y)
                loss.backward()
                optimizer.step()

            for param in model.parameters():
                self.assertTrue(param.grad is not None)
                self.assertGreater(param.grad.norm().item(), 0)
            for param in likelihood.parameters():
                self.assertTrue(param.grad is not None)
                self.assertGreater(param.grad.norm().item(), 0)

            # Set back to eval mode
            model.eval()
            likelihood.eval()
            test_preds = likelihood(model(train_x)).mean.squeeze()
            mean_abs_error = torch.mean(torch.abs(train_y - test_preds) / 2)
            self.assertLess(mean_abs_error.item(), 1e-1)

            self.assertFalse(cg_mock.called)

    def test_predictive_ll_regression_error(self):
        return self.test_regression_error(
            mll_cls=gpytorch.mlls.PredictiveLogLikelihood,
            distribution_cls=gpytorch.variational.MeanFieldVariationalDistribution,
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from typing import Any, Optional

from torch import Tensor

from ..distributions import MultivariateNormal
from .exact_gp import ExactGP

from .gp import GP
from .pyro import _PyroMixin  # This will only contain functions if Pyro is installed


class ApproximateGP(GP, _PyroMixin):
    r"""
    The base class for any Gaussian process latent function to be used in conjunction
    with approximate inference (typically stochastic variational inference).
    This base class can be used to implement most inducing point methods where the
    variational parameters are learned directly.

    :param ~gpytorch.variational._VariationalStrategy variational_strategy: The strategy that determines
        how the model marginalizes over the variational distribution (over inducing points)
        to produce the approximate posterior distribution (over data)

    The :meth:`forward` function should describe how to compute the prior latent distribution
    on a given input. Typically, this will involve a mean and kernel function.
    The result must be a :obj:`~gpytorch.distributions.MultivariateNormal`.

    Example:
        >>> class MyVariationalGP(gpytorch.models.PyroGP):
        >>>     def __init__(self, variational_strategy):
        >>>         super().__init__(variational_strategy)
        >>>         self.mean_module = gpytorch.means.ZeroMean()
        >>>         self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        >>>
        >>>     def forward(self, x):
        >>>         mean = self.mean_module(x)
        >>>         covar = self.covar_module(x)
        >>>         return gpytorch.distributions.MultivariateNormal(mean, covar)
        >>>
        >>> # variational_strategy = ...
        >>> model = MyVariationalGP(variational_strategy)
        >>> likelihood = gpytorch.likelihoods.GaussianLikelihood()
        >>>
        >>> # optimization loop for variational parameters...
        >>>
        >>> # test_x = ...;
        >>> model(test_x)  # Returns the approximate GP latent function at test_x
        >>> likelihood(model(test_x))  # Returns the (approximate) predictive posterior distribution at test_x
    """

    def __init__(self, variational_strategy):
        super().__init__()

        self.variational_strategy = variational_strategy

    def forward(self, x: Tensor):
        raise NotImplementedError

    def pyro_guide(self, input: Tensor, beta: float = 1.0, name_prefix: str = ""):
        r"""
        (For Pyro integration only). The component of a `pyro.guide` that
        corresponds to drawing samples from the latent GP function.

        :param input: The inputs :math:`\mathbf X`.
        :param beta: (default=1.) How much to scale the :math:`\text{KL} [ q(\mathbf f) \Vert p(\mathbf f) ]`
            term by.
        :param name_prefix: (default="") A name prefix to prepend to pyro sample sites.
        """
        return super().pyro_guide(input, beta=beta, name_prefix=name_prefix)

    def pyro_model(self, input: Tensor, beta: float = 1.0, name_prefix: str = "") -> Tensor:
        r"""
        (For Pyro integration only). The component of a `pyro.model` that
        corresponds to drawing samples from the latent GP function.

        :param input: The inputs :math:`\mathbf X`.
        :param beta: (default=1.) How much to scale the :math:`\text{KL} [ q(\mathbf f) \Vert p(\mathbf f) ]`
            term by.
        :param name_prefix: (default="") A name prefix to prepend to pyro sample sites.
        :return: samples from :math:`q(\mathbf f)`
        """
        return super().pyro_model(input, beta=beta, name_prefix=name_prefix)

    def get_fantasy_model(self, inputs: Tensor, targets: Tensor, **kwargs: Any) -> ExactGP:
        r"""
        Returns a new GP model that incorporates the specified inputs and targets as new training data using
        online variational conditioning (OVC).

        This function first casts the inducing points and variational parameters into pseudo-points before
        returning an equivalent ExactGP model with a specialized likelihood.

        .. note::
            If `targets` is a batch (e.g. `b x m`), then the GP returned from this method will be a batch mode GP.
            If `inputs` is of the same (or lesser) dimension as `targets`, then it is assumed that the fantasy points
            are the same for each target batch.

        :param inputs: (`b1 x ... x bk x m x d` or `f x b1 x ... x bk x m x d`) Locations of fantasy
            observations.
        :param targets: (`b1 x ... x bk x m` or `f x b1 x ... x bk x m`) Labels of fantasy observations.
        :return: An `ExactGP` model with `n + m` training examples, where the `m` fantasy examples have been added
            and all test-time caches have been updated.

        Reference: "Conditioning Sparse Variational Gaussian Processes for Online Decision-Making,"
            Maddox, Stanton, Wilson, NeurIPS, '21
            https://papers.nips.cc/paper/2021/hash/325eaeac5bef34937cfdc1bd73034d17-Abstract.html

        """
        return self.variational_strategy.get_fantasy_model(inputs=inputs, targets=targets, **kwargs)

    def __call__(self, inputs: Optional[Tensor], prior: bool = False, **kwargs) -> MultivariateNormal:
        if inputs is not None and inputs.dim() == 1:
            inputs = inputs.unsqueeze(-1)
        return self.variational_strategy(inputs, prior=prior, **kwargs)



#!/usr/bin/env python3

import unittest

import torch
from torch.distributions import HalfCauchy

from gpytorch.priors import HalfCauchyPrior
from gpytorch.test.utils import least_used_cuda_device


class TestHalfCauchyPrior(unittest.TestCase):
    def test_half_cauchy_prior_to_gpu(self):
        if torch.cuda.is_available():
            prior = HalfCauchy(1.0).cuda()
            self.assertEqual(prior.concentration.device.type, "cuda")
            self.assertEqual(prior.rate.device.type, "cuda")

    def test_half_cauchy_prior_validate_args(self):
        with self.assertRaises(ValueError):
            HalfCauchyPrior(-1, validate_args=True)

    def test_half_cauchy_prior_log_prob(self, cuda=False):
        device = torch.device("cuda") if cuda else torch.device("cpu")
        prior = HalfCauchyPrior(0.1)
        dist = HalfCauchy(0.1)

        t = torch.tensor(1.0, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))
        t = torch.tensor([1.5, 0.5], device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))
        t = torch.tensor([[1.0, 0.5], [3.0, 0.25]], device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))

    def test_half_cauchy_prior_log_prob_cuda(self):
        if torch.cuda.is_available():
            with least_used_cuda_device():
                return self.test_half_cauchy_prior_log_prob(cuda=True)

    def test_half_cauchy_prior_log_prob_log_transform(self, cuda=False):
        device = torch.device("cuda") if cuda else torch.device("cpu")
        prior = HalfCauchyPrior(0.1, transform=torch.exp)
        dist = HalfCauchy(0.1)

        t = torch.tensor(0.0, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t.exp())))
        t = torch.tensor([-1, 0.5], device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t.exp())))
        t = torch.tensor([[-1, 0.5], [0.1, -2.0]], device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t.exp())))

    def test_half_cauchy_prior_log_prob_log_transform_cuda(self):
        if torch.cuda.is_available():
            with least_used_cuda_device():
                return self.test_half_cauchy_prior_log_prob_log_transform(cuda=True)

    def test_half_cauchy_prior_batch_log_prob(self, cuda=False):
        device = torch.device("cuda") if cuda else torch.device("cpu")
        prior = HalfCauchyPrior(0.1)
        dist = HalfCauchy(0.1)
        t = torch.ones(2, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))
        t = torch.ones(2, 2, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))

        scale = torch.tensor([0.1, 1.0], device=device)
        prior = HalfCauchyPrior(scale)
        dist = HalfCauchy(scale)
        t = torch.ones(2, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))
        t = torch.ones(2, 2, device=device)
        self.assertTrue(torch.equal(prior.log_prob(t), dist.log_prob(t)))
        with self.assertRaises(ValueError):
            prior.log_prob(torch.ones(3, device=device))
        with self.assertRaises(ValueError):
            prior.log_prob(torch.ones(2, 3, device=device))

    def test_half_cauchy_prior_batch_log_prob_cuda(self):
        if torch.cuda.is_available():
            with least_used_cuda_device():
                return self.test_half_cauchy_prior_batch_log_prob(cuda=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3


def length_safe_zip(*args):
    """Python's `zip(...)` with checks to ensure the arguments have
    the same number of elements.

    NOTE: This converts all args that do not define "__len__" to a list.
    """
    args = [a if hasattr(a, "__len__") else list(a) for a in args]
    if len({len(a) for a in args}) > 1:
        raise ValueError(
            "Expected the lengths of all arguments to be equal. Got lengths "
            f"{[len(a) for a in args]} for args {args}. Did you pass in "
            "fewer inputs than expected?"
        )
    return zip(*args)



#!/usr/bin/env python3

import unittest

import torch

from gpytorch.kernels import PiecewisePolynomialKernel
from gpytorch.test.base_kernel_test_case import BaseKernelTestCase


class TestPiecewisePolynomialKernel(unittest.TestCase, BaseKernelTestCase):
    def create_kernel_no_ard(self, **kwargs):
        return PiecewisePolynomialKernel(q=2, **kwargs)

    def test_computes_piecewise_polynomial_kernel(self):
        a = torch.tensor([[4, 1], [2, 2], [8, 0]], dtype=torch.float)
        b = torch.tensor([[0, 0], [2, 1], [1, 0]], dtype=torch.float)
        kernel = PiecewisePolynomialKernel(q=0)
        kernel.eval()

        def test_r(a, b):
            return torch.cdist(a, b)

        def test_get_cov(r, j, q):
            if q == 0:
                return 1
            if q == 1:
                return (j + 1) * r + 1
            if q == 2:
                return 1 + (j + 2) * r + ((j**2 + 4 * j + 3) / 3.0) * r**2
            if q == 3:
                return (
                    1
                    + (j + 3) * r
                    + ((6 * j**2 + 36 * j + 45) / 15.0) * r**2
                    + ((j**3 + 9 * j**2 + 23 * j + 15) / 15.0) * r**3
                )

        def test_fmax(r, j, q):
            return torch.max(torch.tensor(0.0), 1 - r).pow(j + q)

        actual = torch.zeros(3, 3)
        j = torch.floor(a / 2.0).shape[-1] + kernel.q + 1
        r = test_r(a, b)
        actual = test_fmax(r, j, kernel.q) * test_get_cov(r, j, kernel.q)
        res = kernel(a, b).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)

        # diag
        actual = actual.diagonal(dim1=-1, dim2=-2)
        res = kernel(a, b).diagonal(dim1=-1, dim2=-2)
        self.assertLess(torch.norm(res - actual), 1e-5)

        # batch_dims
        actual = torch.zeros(2, 3, 3)
        for i in range(2):
            actual[i] = kernel(a[:, i].unsqueeze(-1), b[:, i].unsqueeze(-1)).to_dense()

        res = kernel(a, b, last_dim_is_batch=True).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)

        # batch_dims + diag
        res = kernel(a, b, last_dim_is_batch=True).diagonal(dim1=-1, dim2=-2)
        actual = torch.cat([actual[i].diagonal(dim1=-1, dim2=-2).unsqueeze(0) for i in range(actual.size(0))])
        self.assertLess(torch.norm(res - actual), 1e-5)

    def test_piecewise_polynomial_kernel_batch(self):
        a = torch.tensor([[4, 2, 8], [1, 2, 3]], dtype=torch.float).view(2, 3, 1)
        b = torch.tensor([[0, 2, 1], [-1, 2, 0]], dtype=torch.float).view(2, 3, 1)
        kernel = PiecewisePolynomialKernel(q=0, batch_shape=torch.Size([2]))
        kernel.eval()

        def test_r(a, b):
            return torch.cdist(a, b)

        def test_get_cov(r, j, q):
            if q == 0:
                return 1
            if q == 1:
                return (j + 1) * r + 1
            if q == 2:
                return 1 + (j + 2) * r + ((j**2 + 4 * j + 3) / 3.0) * r**2
            if q == 3:
                return (
                    1
                    + (j + 3) * r
                    + ((6 * j**2 + 36 * j + 45) / 15.0) * r**2
                    + ((j**3 + 9 * j**2 + 23 * j + 15) / 15.0) * r**3
                )

        def test_fmax(r, j, q):
            return torch.max(torch.tensor(0.0), 1 - r).pow(j + q)

        actual = torch.zeros(3, 3)
        j = torch.floor(a / 2.0).shape[-1] + kernel.q + 1
        r = test_r(a, b)
        actual = test_fmax(r, j, kernel.q) * test_get_cov(r, j, kernel.q)
        res = kernel(a, b).to_dense()
        self.assertLess(torch.norm(res - actual), 1e-5)


if __name__ == "__main__":
    unittest.main()



import warnings
from typing import Any, Tuple, Union

import torch
from linear_operator import LinearOperator
from torch import Tensor

from ... import settings
from ..kernel import Kernel

try:
    import pykeops  # noqa F401
    from pykeops.torch import LazyTensor

    _Anysor = Union[Tensor, LazyTensor]

    def _lazify_and_expand_inputs(
        x1: Tensor, x2: Tensor
    ) -> Tuple[Union[Tensor, LazyTensor], Union[Tensor, LazyTensor]]:
        r"""
        Potentially wrap inputs x1 and x2 as KeOps LazyTensors,
        depending on whether or not we want to use KeOps under the hood or not.
        """
        x1_ = x1[..., :, None, :]
        x2_ = x2[..., None, :, :]
        if _use_keops(x1, x2):
            res = LazyTensor(x1_), LazyTensor(x2_)
            return res
        return x1_, x2_

    def _use_keops(x1: Tensor, x2: Tensor) -> bool:
        r"""
        Determine whether or not to use KeOps under the hood
        This largely depends on the size of the kernel matrix

        There are situations where we do not want the KeOps linear operator to use KeOps under the hood.
        See https://github.com/cornellius-gp/gpytorch/pull/1319
        """
        return (
            settings.use_keops.on()
            and x1.size(-2) >= settings.max_cholesky_size.value()
            and x2.size(-2) >= settings.max_cholesky_size.value()
        )

    class KeOpsKernel(Kernel):
        def __call__(self, *args: Any, **kwargs: Any) -> Union[LinearOperator, Tensor, LazyTensor]:
            # Hotfix for zero gradients. See https://github.com/cornellius-gp/gpytorch/issues/1543
            args = [arg.contiguous() if torch.is_tensor(arg) else arg for arg in args]
            kwargs = {k: v.contiguous() if torch.is_tensor(v) else v for k, v in kwargs.items()}
            return super().__call__(*args, **kwargs)

except ImportError:

    _Anysor = Tensor

    def _lazify_and_expand_inputs(x1: Tensor, x2: Tensor) -> Tuple[Tensor, Tensor]:
        x1_ = x1[..., :, None, :]
        x2_ = x2[..., None, :, :]
        return x1_, x2_

    def _use_keops(x1: Tensor, x2: Tensor) -> bool:
        return False

    class KeOpsKernel(Kernel):
        def __call__(self, *args: Any, **kwargs: Any) -> Union[LinearOperator, Tensor]:
            warnings.warn(
                "KeOps is not installed. " f"{type(self)} will revert to the the non-keops version of this kernel.",
                RuntimeWarning,
            )
            return super().__call__(*args, **kwargs)



#!/usr/bin/env python3

import math
import pickle
import unittest
import warnings
from collections import OrderedDict

import torch

import gpytorch
from gpytorch.constraints import GreaterThan
from gpytorch.means import ConstantMean
from gpytorch.priors import NormalPrior
from gpytorch.test.base_mean_test_case import BaseMeanTestCase
from gpytorch.utils.warnings import OldVersionWarning


# Test class for loading models that have state dicts with the old ConstantMean parameter names
class _GPModel(gpytorch.models.ExactGP):
    def __init__(self, mean_module):
        train_x = torch.randn(10, 3)
        train_y = torch.randn(10)
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = mean_module


class TestConstantMean(BaseMeanTestCase, unittest.TestCase):
    batch_shape = None

    def create_mean(self, prior=None, constraint=None):
        return ConstantMean(
            constant_prior=prior,
            constant_constraint=constraint,
            batch_shape=(self.__class__.batch_shape or torch.Size([])),
        )

    def test_prior(self):
        if self.batch_shape is None:
            prior = NormalPrior(0.0, 1.0)
        else:
            prior = NormalPrior(torch.zeros(self.batch_shape), torch.ones(self.batch_shape))
        mean = self.create_mean(prior=prior)
        self.assertEqual(mean.mean_prior, prior)
        pickle.loads(pickle.dumps(mean))  # Should be able to pickle and unpickle with a prior
        value = prior.sample()
        mean._constant_closure(mean, value)
        self.assertTrue(torch.equal(mean.constant.data, value.reshape(mean.constant.data.shape)))

    def test_constraint(self):
        mean = self.create_mean()
        self.assertAllClose(mean.constant, torch.zeros(mean.constant.shape))

        constraint = GreaterThan(1.5)
        mean = self.create_mean(constraint=constraint)
        self.assertTrue(torch.all(mean.constant >= 1.5))
        mean.constant = torch.full(self.__class__.batch_shape or torch.Size([]), fill_value=1.65)
        self.assertAllClose(mean.constant, torch.tensor(1.65).expand(mean.constant.shape))

    def test_loading_old_module(self):
        batch_shape = self.__class__.batch_shape or torch.Size([])
        constant = torch.randn(batch_shape)
        mean = self.create_mean()
        model = _GPModel(mean)

        old_state_dict = OrderedDict(
            [
                ("likelihood.noise_covar.raw_noise", torch.tensor([0.0])),
                ("likelihood.noise_covar.raw_noise_constraint.lower_bound", torch.tensor(1.0000e-04)),
                ("likelihood.noise_covar.raw_noise_constraint.upper_bound", torch.tensor(math.inf)),
                ("mean_module.constant", constant.unsqueeze(-1)),
            ]
        )
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter("always", OldVersionWarning)
            model.load_state_dict(old_state_dict)
            self.assertTrue(any(issubclass(w.category, OldVersionWarning) for w in ws))
            self.assertEqual(model.mean_module.constant.data, constant)

        new_state_dict = OrderedDict(
            [
                ("likelihood.noise_covar.raw_noise", torch.tensor([0.0])),
                ("likelihood.noise_covar.raw_noise_constraint.lower_bound", torch.tensor(1.0000e-04)),
                ("likelihood.noise_covar.raw_noise_constraint.upper_bound", torch.tensor(math.inf)),
                ("mean_module.raw_constant", constant),
            ]
        )
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter("always", OldVersionWarning)
            model.load_state_dict(new_state_dict)
            self.assertFalse(any(issubclass(w.category, OldVersionWarning) for w in ws))


class TestConstantMeanBatch(TestConstantMean, unittest.TestCase):
    batch_shape = torch.Size([3])


class TestConstantMeanMultiBatch(TestConstantMean, unittest.TestCase):
    batch_shape = torch.Size([2, 3])



#!/usr/bin/env python3

import unittest

import torch

import gpytorch
from gpytorch.likelihoods import HeteroskedasticNoise


class NumericallyUnstableModelExample(gpytorch.models.GP):
    def __init__(self):
        super(NumericallyUnstableModelExample, self).__init__()
        self.fail_arithmetic = False

    def train(self, mode=True):
        if mode:
            self.fail_arithmetic = False  # reset on .train()
        super().train(mode=mode)

    def forward(self, x):
        if self.fail_arithmetic:
            raise ArithmeticError()
        return gpytorch.distributions.MultivariateNormal(torch.tensor([-3.0]), torch.tensor([[2.0]]))


class TestNoiseModels(unittest.TestCase):
    def test_heteroskedasticnoise_error(self):
        noise_model = NumericallyUnstableModelExample().to(torch.double)
        likelihood = HeteroskedasticNoise(noise_model)
        self.assertEqual(noise_model.training, True)
        self.assertEqual(likelihood.training, True)
        noise_model.fail_arithmetic = True
        test_x = torch.tensor([[3.0]])
        with self.assertRaises(ArithmeticError):
            likelihood(test_x)
        self.assertEqual(likelihood.training, True)
        likelihood(test_x)



import unittest

import torch

import gpytorch
from gpytorch.kernels import MaternKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ConstantMean, LinearMean
from gpytorch.models.deep_gps.dspp import DSPP, DSPPLayer
from gpytorch.test.base_test_case import BaseTestCase
from gpytorch.variational import MeanFieldVariationalDistribution, VariationalStrategy


train_n = 20
train_x = torch.linspace(0, 1, train_n).unsqueeze(-1)
train_y = torch.sin(2 * train_x[..., 0]) + torch.randn(train_n).mul_(0.01)
test_n = 10
test_x = torch.linspace(-0.2, 1.2, test_n).unsqueeze(-1)
test_y = torch.sin(2 * test_x[..., 0])

num_inducing_pts = 10  # Number of inducing points in each hidden layer
num_epochs = 100  # Number of epochs to train for
initial_lr = 0.1  # Initial learning rate
hidden_dim = 3  # Number of GPs (i.e., the width) in the hidden layer.
num_quadrature_sites = 8  # Number of quadrature sites (see paper for a description of this.


class DSPPHiddenLayer(DSPPLayer):
    def __init__(self, input_dims, output_dims, num_inducing=300, inducing_points=None, mean_type="constant", Q=8):
        if output_dims is None:
            # An output_dims of None implies there is only one GP in this layer
            # (e.g., the last layer for univariate regression).
            inducing_points = torch.randn(num_inducing, input_dims)
        else:
            inducing_points = torch.randn(output_dims, num_inducing, input_dims)

        # Let's use mean field / diagonal covariance structure.
        variational_distribution = MeanFieldVariationalDistribution(
            num_inducing_points=num_inducing,
            batch_shape=torch.Size([output_dims]) if output_dims is not None else torch.Size([]),
        )

        # Standard variational inference.
        variational_strategy = VariationalStrategy(
            self, inducing_points, variational_distribution, learn_inducing_locations=True
        )

        batch_shape = torch.Size([]) if output_dims is None else torch.Size([output_dims])
        super().__init__(variational_strategy, input_dims, output_dims, Q)

        if mean_type == "constant":
            self.mean_module = ConstantMean(batch_shape=batch_shape)
        elif mean_type == "linear":
            self.mean_module = LinearMean(input_dims, batch_shape=batch_shape)

        self.covar_module = ScaleKernel(
            MaternKernel(batch_shape=batch_shape, ard_num_dims=input_dims), batch_shape=batch_shape, ard_num_dims=None
        )

    def forward(self, x, mean_input=None, **kwargs):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class ThreeLayerDSPP(DSPP):
    def __init__(self, train_x_shape, num_inducing, hidden_dim=3, Q=3):
        hidden_layer1 = DSPPHiddenLayer(
            input_dims=train_x_shape[-1],
            output_dims=hidden_dim,
            num_inducing=num_inducing,
            mean_type="linear",
            Q=Q,
        )
        hidden_layer2 = DSPPHiddenLayer(
            input_dims=hidden_layer1.output_dims,
            output_dims=hidden_dim,
            mean_type="linear",
            num_inducing=num_inducing,
            Q=Q,
        )
        last_layer = DSPPHiddenLayer(
            input_dims=hidden_layer2.output_dims,
            output_dims=None,
            mean_type="constant",
            num_inducing=num_inducing,
            Q=Q,
        )

        likelihood = GaussianLikelihood()

        super().__init__(Q)
        self.likelihood = likelihood
        self.last_layer = last_layer
        self.hidden_layer2 = hidden_layer2
        self.hidden_layer1 = hidden_layer1

    def forward(self, inputs, **kwargs):
        hidden_rep1 = self.hidden_layer1(inputs, **kwargs)
        hidden_rep2 = self.hidden_layer2(hidden_rep1, **kwargs)
        output = self.last_layer(hidden_rep2, expand_for_quadgrid=False, **kwargs)
        return output


class TestSGPRRegression(unittest.TestCase, BaseTestCase):
    seed = 0

    def test_dspp(self):
        model = ThreeLayerDSPP(
            train_x.shape, num_inducing=num_inducing_pts, hidden_dim=hidden_dim, Q=num_quadrature_sites
        )

        model.train()

        adam = torch.optim.Adam([{"params": model.parameters()}], lr=initial_lr, betas=(0.9, 0.999))
        objective = gpytorch.mlls.DeepPredictiveLogLikelihood(model.likelihood, model, num_data=train_n, beta=0.05)

        epochs_iter = range(num_epochs)
        losses = []

        for i in epochs_iter:
            adam.zero_grad()
            output = model(train_x)
            loss = -objective(output, train_y)
            loss.backward()
            adam.step()
            losses.append(loss.item())

        model.eval()
        preds = model.likelihood(model(test_x))
        self.assertLess((preds.mean - test_y).abs().mean().item(), 0.5)
        self.assertLess(preds.variance.mean().item(), 0.5)



from typing import Optional, Union

import torch
from torch import nn, Tensor

from gpytorch.constraints.constraints import Interval, Positive
from gpytorch.kernels.kernel import Kernel
from gpytorch.priors.prior import Prior


EMPTY_SIZE = torch.Size([])


class HammingIMQKernel(Kernel):
    r"""
    Computes a covariance matrix based on the inverse multiquadratic Hamming kernel
    between inputs :math:`\mathbf{x_1}` and :math:`\mathbf{x_2}`:

    .. math::
       \begin{equation*}
            k_{\text{H-IMQ}}(\mathbf{x_1}, \mathbf{x_2}) =
            \left( \frac{1 + \alpha}{\alpha + d_{\text{Hamming}}(x1, x2)} \right)^\beta
       \end{equation*}
    where :math:`\alpha` and :math:`\beta` are strictly positive scale parameters.
    This kernel was proposed in `Biological Sequence Kernels with Guaranteed Flexibility`.
    See http://arxiv.org/abs/2304.03775 for more details.

    This kernel is meant to be used for fixed-length one-hot encoded discrete sequences.
    Because GPyTorch is particular about dimensions, the one-hot sequence encoding should be flattened
    to a vector with length :math:`T \times V`, where :math:`T` is the sequence length and :math:`V` is the
    vocabulary size.

    :param vocab_size: The size of the vocabulary.
    :param batch_shape: Set this if you want a separate kernel hyperparameters for each batch of input
        data. It should be :math:`B_1 \times \ldots \times B_k` if :math:`\mathbf{x_1}` is
        a :math:`B_1 \times \ldots \times B_k \times N \times D` tensor.
    :param alpha_prior: Set this if you want to apply a prior to the
        alpha parameter.
    :param: alpha_constraint: Set this if you want to apply a constraint
        to the alpha parameter. If None is passed, the default is `Positive()`.
    :param beta_prior: Set this if you want to apply a prior to the
        beta parameter.
    :param beta_constraint: Set this if you want to apply a constraint
        to the beta parameter. If None is passed, the default is `Positive()`.

    Example:
        >>> vocab_size = 8
        >>> x_cat = torch.tensor([[7, 7, 7, 7], [5, 7, 3, 4]])  # batch_size x seq_length
        >>> x_one_hot = F.one_hot(x_cat, num_classes=vocab_size)  # batch_size x seq_length x vocab_size
        >>> x_flat = x_one_hot.view(*x_cat.shape[:-1], -1)  # batch_size x (seq_length * vocab_size)
        >>> covar_module = gpytorch.kernels.HammingIMQKernel(vocab_size=vocab_size)
        >>> covar = covar_module(x_flat)  # Output: LinearOperator of size (2 x 2)
    """

    def __init__(
        self,
        vocab_size: int,
        batch_shape: torch.Size = EMPTY_SIZE,
        alpha_prior: Optional[Prior] = None,
        alpha_constraint: Optional[Interval] = None,
        beta_prior: Optional[Prior] = None,
        beta_constraint: Optional[Interval] = None,
    ):
        super().__init__(batch_shape=batch_shape)
        self.vocab_size = vocab_size
        # add alpha (scale) parameter
        alpha_constraint = Positive() if alpha_constraint is None else alpha_constraint
        self.register_parameter(
            name="raw_alpha",
            parameter=nn.Parameter(torch.zeros(*self.batch_shape, 1)),
        )
        if alpha_prior is not None:
            self.register_prior("alpha_prior", alpha_prior, self._alpha_param, self._alpha_closure)
        self.register_constraint("raw_alpha", alpha_constraint)

        # add beta parameter
        beta_constraint = Positive() if beta_constraint is None else beta_constraint
        self.register_parameter(
            name="raw_beta",
            parameter=nn.Parameter(torch.zeros(*self.batch_shape, 1)),
        )
        if beta_prior is not None:
            self.register_prior("beta_prior", beta_prior, self._beta_param, self._beta_closure)
        self.register_constraint("raw_beta", beta_constraint)

    @property
    def alpha(self) -> Tensor:
        return self.raw_alpha_constraint.transform(self.raw_alpha)

    @alpha.setter
    def alpha(self, value: Tensor):
        self._set_alpha(value)

    def _alpha_param(self, m: Kernel) -> Tensor:
        # Used by the alpha_prior
        return m.alpha

    def _alpha_closure(self, m: Kernel, v: Union[Tensor, float]) -> None:
        # Used by the alpha_prior
        m._set_alpha(v)

    def _set_alpha(self, value: Union[Tensor, float]) -> None:
        # Used by the alpha_prior
        if not isinstance(value, Tensor):
            value = torch.as_tensor(value).to(self.raw_alpha)
        self.initialize(raw_alpha=self.raw_alpha_constraint.inverse_transform(value))

    @property
    def beta(self) -> Tensor:
        return self.raw_beta_constraint.transform(self.raw_beta)

    @beta.setter
    def beta(self, value: Tensor):
        self._set_beta(value)

    def _beta_param(self, m: Kernel) -> Tensor:
        # Used by the beta_prior
        return m.beta

    def _beta_closure(self, m: Kernel, v: Union[Tensor, float]) -> None:
        # Used by the beta_prior
        m._set_beta(v)

    def _set_beta(self, value: Union[Tensor, float]) -> None:
        # Used by the beta_prior
        if not isinstance(value, Tensor):
            value = torch.as_tensor(value).to(self.raw_beta)
        self.initialize(raw_beta=self.raw_beta_constraint.inverse_transform(value))

    def _imq(self, dist: Tensor) -> Tensor:
        return ((1 + self.alpha) / (self.alpha + dist)).pow(self.beta)

    def forward(self, x1: Tensor, x2: Tensor, diag: bool = False, **params):
        # GPyTorch is pretty particular about dimensions so we need to unflatten the one-hot encoding
        x1 = x1.view(*x1.shape[:-1], -1, self.vocab_size)
        x2 = x2.view(*x2.shape[:-1], -1, self.vocab_size)

        x1_eq_x2 = torch.equal(x1, x2)

        if diag:
            if x1_eq_x2:
                res = ((1 + self.alpha) / self.alpha).pow(self.beta)
                skip_dims = [-1] * len(self.batch_shape)
                return res.expand(*skip_dims, x1.size(-3))
            else:
                dist = x1.size(-2) - (x1 * x2).sum(dim=(-1, -2))
                return self._imq(dist)

        else:
            dist = hamming_dist(x1, x2, x1_eq_x2)

        return self._imq(dist)


def hamming_dist(x1: Tensor, x2: Tensor, x1_eq_x2: bool) -> Tensor:
    res = x1.size(-2) - (x1.unsqueeze(-3) * x2.unsqueeze(-4)).sum(dim=(-1, -2))
    if x1_eq_x2 and not x1.requires_grad and not x2.requires_grad:
        res.diagonal(dim1=-2, dim2=-1).fill_(0)
    # Zero out negative values
    return res.clamp_min_(0)



#!/usr/bin/env python3

import unittest
from math import pi

import torch
from torch import optim

import gpytorch
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import ScaleKernel
from gpytorch.kernels.keops import RBFKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ConstantMean
from gpytorch.test.base_test_case import BaseTestCase


# Simple training data: let's try to learn a sine function
train_x = torch.randn(300, 2)
train_y = torch.sin(train_x[..., 0] * (2 * pi) + train_x[..., 1])
train_y = train_y + torch.randn_like(train_y).mul(0.001)

test_x = torch.randn(50, 2)
test_y = torch.sin(test_x[..., 0] * (2 * pi) + test_x[..., 1])


class KeOpsGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(RBFKernel(ard_num_dims=2))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class TestKeOpsGPRegression(BaseTestCase, unittest.TestCase):
    seed = 4

    def test_keops_gp_mean_abs_error(self):
        try:
            import pykeops  # noqa
        except ImportError:
            return

        likelihood = GaussianLikelihood()
        gp_model = KeOpsGPModel(train_x, train_y, likelihood)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, gp_model)

        # Optimize the model
        gp_model.train()
        likelihood.train()
        optimizer = optim.Adam(list(gp_model.parameters()), lr=0.1)
        optimizer.n_iter = 0

        with gpytorch.settings.max_cholesky_size(0):  # Ensure that we're using KeOps
            for i in range(25):
                optimizer.zero_grad()
                output = gp_model(train_x)
                loss = -mll(output, train_y)
                loss.backward()
                optimizer.step()

                if i == 0:
                    for param in gp_model.parameters():
                        self.assertTrue(param.grad is not None)

            # Test the model
            with torch.no_grad():
                gp_model.eval()
                likelihood.eval()
                test_preds = likelihood(gp_model(test_x)).mean
                mean_abs_error = torch.mean(torch.abs(test_y - test_preds))

        self.assertLess(mean_abs_error.squeeze().item(), 0.15)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from .bernoulli_likelihood import BernoulliLikelihood
from .beta_likelihood import BetaLikelihood
from .gaussian_likelihood import (
    _GaussianLikelihoodBase,
    DirichletClassificationLikelihood,
    FixedNoiseGaussianLikelihood,
    GaussianLikelihood,
    GaussianLikelihoodWithMissingObs,
)
from .hadamard_gaussian_likelihood import HadamardGaussianLikelihood
from .laplace_likelihood import LaplaceLikelihood
from .likelihood import _OneDimensionalLikelihood, Likelihood
from .likelihood_list import LikelihoodList
from .multitask_gaussian_likelihood import _MultitaskGaussianLikelihoodBase, MultitaskGaussianLikelihood
from .noise_models import HeteroskedasticNoise
from .softmax_likelihood import SoftmaxLikelihood
from .student_t_likelihood import StudentTLikelihood

__all__ = [
    "_GaussianLikelihoodBase",
    "_OneDimensionalLikelihood",
    "_MultitaskGaussianLikelihoodBase",
    "BernoulliLikelihood",
    "BetaLikelihood",
    "DirichletClassificationLikelihood",
    "FixedNoiseGaussianLikelihood",
    "GaussianLikelihood",
    "GaussianLikelihoodWithMissingObs",
    "HadamardGaussianLikelihood",
    "HeteroskedasticNoise",
    "LaplaceLikelihood",
    "Likelihood",
    "LikelihoodList",
    "MultitaskGaussianLikelihood",
    "SoftmaxLikelihood",
    "StudentTLikelihood",
]



#!/usr/bin/env python3

import torch
from linear_operator import LinearOperator, to_linear_operator
from linear_operator.operators import (
    BlockDiagLinearOperator,
    BlockInterleavedLinearOperator,
    CatLinearOperator,
    DiagLinearOperator,
)

from .multivariate_normal import MultivariateNormal


class MultitaskMultivariateNormal(MultivariateNormal):
    """
    Constructs a multi-output multivariate Normal random variable, based on mean and covariance
    Can be multi-output multivariate, or a batch of multi-output multivariate Normal

    Passing a matrix mean corresponds to a multi-output multivariate Normal
    Passing a matrix mean corresponds to a batch of multivariate Normals

    :param torch.Tensor mean:  An `n x t` or batch `b x n x t` matrix of means for the MVN distribution.
    :param ~linear_operator.operators.LinearOperator covar: An `... x NT x NT` (batch) matrix.
        covariance matrix of MVN distribution.
    :param bool validate_args: (default=False) If True, validate `mean` and `covariance_matrix` arguments.
    :param bool interleaved: (default=True) If True, covariance matrix is interpreted as block-diagonal w.r.t.
        inter-task covariances for each observation. If False, it is interpreted as block-diagonal
        w.r.t. inter-observation covariance for each task.
    """

    def __init__(self, mean, covariance_matrix, validate_args=False, interleaved=True):
        if not torch.is_tensor(mean) and not isinstance(mean, LinearOperator):
            raise RuntimeError("The mean of a MultitaskMultivariateNormal must be a Tensor or LinearOperator")

        if not torch.is_tensor(covariance_matrix) and not isinstance(covariance_matrix, LinearOperator):
            raise RuntimeError("The covariance of a MultitaskMultivariateNormal must be a Tensor or LinearOperator")

        if mean.dim() < 2:
            raise RuntimeError("mean should be a matrix or a batch matrix (batch mode)")

        # Ensure that shapes are broadcasted appropriately across the mean and covariance
        # Means can have singleton dimensions for either the `n` or `t` dimensions
        batch_shape = torch.broadcast_shapes(mean.shape[:-2], covariance_matrix.shape[:-2])
        if mean.shape[-2:].numel() != covariance_matrix.size(-1):
            if covariance_matrix.size(-1) % mean.shape[-2:].numel():
                raise RuntimeError(
                    f"mean shape {mean.shape} is incompatible with covariance shape {covariance_matrix.shape}"
                )
            elif mean.size(-2) == 1:
                mean = mean.expand(*batch_shape, covariance_matrix.size(-1) // mean.size(-1), mean.size(-1))
            elif mean.size(-1) == 1:
                mean = mean.expand(*batch_shape, mean.size(-2), covariance_matrix.size(-2) // mean.size(-2))
            else:
                raise RuntimeError(
                    f"mean shape {mean.shape} is incompatible with covariance shape {covariance_matrix.shape}"
                )
        else:
            mean = mean.expand(*batch_shape, *mean.shape[-2:])

        self._output_shape = mean.shape
        # TODO: Instead of transpose / view operations, use a PermutationLinearOperator (see #539)
        # to handle interleaving
        self._interleaved = interleaved
        if self._interleaved:
            mean_mvn = mean.reshape(*mean.shape[:-2], -1)
        else:
            mean_mvn = mean.transpose(-1, -2).reshape(*mean.shape[:-2], -1)
        super().__init__(mean=mean_mvn, covariance_matrix=covariance_matrix, validate_args=validate_args)

    @property
    def base_sample_shape(self):
        """
        Returns the shape of a base sample (without batching) that is used to
        generate a single sample.
        """
        base_sample_shape = self.event_shape
        return base_sample_shape

    @property
    def event_shape(self):
        return self._output_shape[-2:]

    @classmethod
    def from_batch_mvn(cls, batch_mvn, task_dim=-1):
        """
        Reinterprate a batch of multivariate normal distributions as an (independent) multitask multivariate normal
        distribution.

        :param ~gpytorch.distributions.MultivariateNormal batch_mvn: The base MVN distribution.
            (This distribution should have at least one batch dimension).
        :param int task_dim: Which batch dimension should be interpreted as the dimension for the independent tasks.
        :returns: the independent multitask distribution
        :rtype: gpytorch.distributions.MultitaskMultivariateNormal

        Example:
            >>> # model is a gpytorch.models.VariationalGP
            >>> # likelihood is a gpytorch.likelihoods.Likelihood
            >>> mean = torch.randn(4, 2, 3)
            >>> covar_factor = torch.randn(4, 2, 3, 3)
            >>> covar = covar_factor @ covar_factor.transpose(-1, -2)
            >>> mvn = gpytorch.distributions.MultivariateNormal(mean, covar)
            >>> print(mvn.event_shape, mvn.batch_shape)
            >>> # torch.Size([3]), torch.Size([4, 2])
            >>>
            >>> mmvn = MultitaskMultivariateNormal.from_batch_mvn(mvn, task_dim=-1)
            >>> print(mmvn.event_shape, mmvn.batch_shape)
            >>> # torch.Size([3, 2]), torch.Size([4])
        """
        orig_task_dim = task_dim
        task_dim = task_dim if task_dim >= 0 else (len(batch_mvn.batch_shape) + task_dim)
        if task_dim < 0 or task_dim > len(batch_mvn.batch_shape):
            raise ValueError(
                f"task_dim of {orig_task_dim} is incompatible with MVN batch shape of {batch_mvn.batch_shape}"
            )

        num_dim = batch_mvn.mean.dim()
        res = cls(
            mean=batch_mvn.mean.permute(*range(0, task_dim), *range(task_dim + 1, num_dim), task_dim),
            covariance_matrix=BlockInterleavedLinearOperator(batch_mvn.lazy_covariance_matrix, block_dim=task_dim),
        )
        return res

    @classmethod
    def from_independent_mvns(cls, mvns):
        """
        Convert an iterable of MVNs into a :obj:`~gpytorch.distributions.MultitaskMultivariateNormal`.
        The resulting distribution will have ``len(mvns)`` tasks, and the tasks will be independent.

        :param ~gpytorch.distributions.MultitaskNormal mvn: The base MVN distributions.
        :returns: the independent multitask distribution
        :rtype: gpytorch.distributions.MultitaskMultivariateNormal

        Example:
            >>> # model is a gpytorch.models.VariationalGP
            >>> # likelihood is a gpytorch.likelihoods.Likelihood
            >>> mean = torch.randn(4, 3)
            >>> covar_factor = torch.randn(4, 3, 3)
            >>> covar = covar_factor @ covar_factor.transpose(-1, -2)
            >>> mvn1 = gpytorch.distributions.MultivariateNormal(mean, covar)
            >>>
            >>> mean = torch.randn(4, 3)
            >>> covar_factor = torch.randn(4, 3, 3)
            >>> covar = covar_factor @ covar_factor.transpose(-1, -2)
            >>> mvn2 = gpytorch.distributions.MultivariateNormal(mean, covar)
            >>>
            >>> mmvn = MultitaskMultivariateNormal.from_independent_mvns([mvn1, mvn2])
            >>> print(mmvn.event_shape, mmvn.batch_shape)
            >>> # torch.Size([3, 2]), torch.Size([4])
        """
        if len(mvns) < 2:
            raise ValueError("Must provide at least 2 MVNs to form a MultitaskMultivariateNormal")
        if any(isinstance(mvn, MultitaskMultivariateNormal) for mvn in mvns):
            raise ValueError("Cannot accept MultitaskMultivariateNormals")
        if not all(m.batch_shape == mvns[0].batch_shape for m in mvns[1:]):
            batch_shape = torch.broadcast_shapes(*(m.batch_shape for m in mvns))
            mvns = [mvn.expand(batch_shape) for mvn in mvns]
        if not all(m.event_shape == mvns[0].event_shape for m in mvns[1:]):
            raise ValueError("All MultivariateNormals must have the same event shape")
        mean = torch.stack([mvn.mean for mvn in mvns], -1)
        # TODO: To do the following efficiently, we don't want to evaluate the
        # covariance matrices. Instead, we want to use the lazies directly in the
        # BlockDiagLinearOperator. This will require implementing a new BatchLinearOperator:

        # https://github.com/cornellius-gp/gpytorch/issues/468
        covar_blocks_lazy = CatLinearOperator(
            *[mvn.lazy_covariance_matrix.unsqueeze(0) for mvn in mvns], dim=0, output_device=mean.device
        )
        covar_lazy = BlockDiagLinearOperator(covar_blocks_lazy, block_dim=0)
        return cls(mean=mean, covariance_matrix=covar_lazy, interleaved=False)

    @classmethod
    def from_repeated_mvn(cls, mvn, num_tasks):
        """
        Convert a single MVN into a :obj:`~gpytorch.distributions.MultitaskMultivariateNormal`,
        where each task shares the same mean and covariance.

        :param ~gpytorch.distributions.MultitaskNormal mvn: The base MVN distribution.
        :param int num_tasks: How many tasks to create.
        :returns: the independent multitask distribution
        :rtype: gpytorch.distributions.MultitaskMultivariateNormal

        Example:
            >>> # model is a gpytorch.models.VariationalGP
            >>> # likelihood is a gpytorch.likelihoods.Likelihood
            >>> mean = torch.randn(4, 3)
            >>> covar_factor = torch.randn(4, 3, 3)
            >>> covar = covar_factor @ covar_factor.transpose(-1, -2)
            >>> mvn = gpytorch.distributions.MultivariateNormal(mean, covar)
            >>> print(mvn.event_shape, mvn.batch_shape)
            >>> # torch.Size([3]), torch.Size([4])
            >>>
            >>> mmvn = MultitaskMultivariateNormal.from_repeated_mvn(mvn, num_tasks=2)
            >>> print(mmvn.event_shape, mmvn.batch_shape)
            >>> # torch.Size([3, 2]), torch.Size([4])
        """
        return cls.from_batch_mvn(mvn.expand(torch.Size([num_tasks]) + mvn.batch_shape), task_dim=0)

    def expand(self, batch_size):
        new_mean = self.mean.expand(torch.Size(batch_size) + self.mean.shape[-2:])
        new_covar = self._covar.expand(torch.Size(batch_size) + self._covar.shape[-2:])
        res = self.__class__(new_mean, new_covar, interleaved=self._interleaved)
        return res

    def get_base_samples(self, sample_shape=torch.Size()):
        base_samples = super().get_base_samples(sample_shape)
        if not self._interleaved:
            # flip shape of last two dimensions
            new_shape = sample_shape + self._output_shape[:-2] + self._output_shape[:-3:-1]
            return base_samples.view(new_shape).transpose(-1, -2).contiguous()
        return base_samples.view(*sample_shape, *self._output_shape)

    def log_prob(self, value):
        if not self._interleaved:
            # flip shape of last two dimensions
            new_shape = value.shape[:-2] + value.shape[:-3:-1]
            value = value.view(new_shape).transpose(-1, -2).contiguous()
        return super().log_prob(value.reshape(*value.shape[:-2], -1))

    @property
    def mean(self):
        mean = super().mean
        if not self._interleaved:
            # flip shape of last two dimensions
            new_shape = self._output_shape[:-2] + self._output_shape[:-3:-1]
            return mean.view(new_shape).transpose(-1, -2).contiguous()
        return mean.view(self._output_shape)

    @property
    def num_tasks(self):
        return self._output_shape[-1]

    def rsample(self, sample_shape=torch.Size(), base_samples=None):
        if base_samples is not None:
            # Make sure that the base samples agree with the distribution
            mean_shape = self.mean.shape
            base_sample_shape = base_samples.shape[-self.mean.ndimension() :]
            if mean_shape != base_sample_shape:
                raise RuntimeError(
                    "The shape of base_samples (minus sample shape dimensions) should agree with the shape "
                    "of self.mean. Expected ...{} but got {}".format(mean_shape, base_sample_shape)
                )
            sample_shape = base_samples.shape[: -self.mean.ndimension()]
            base_samples = base_samples.view(*sample_shape, *self.loc.shape)

        samples = super().rsample(sample_shape=sample_shape, base_samples=base_samples)
        if not self._interleaved:
            # flip shape of last two dimensions
            new_shape = sample_shape + self._output_shape[:-2] + self._output_shape[:-3:-1]
            return samples.view(new_shape).transpose(-1, -2).contiguous()
        return samples.view(sample_shape + self._output_shape)

    def to_data_independent_dist(self, jitter_val=1e-4):
        """
        Convert a multitask MVN into a batched (non-multitask) MVNs
        The result retains the intertask covariances, but gets rid of the inter-data covariances.
        The resulting distribution will have ``len(mvns)`` tasks, and the tasks will be independent.

        :returns: the bached data-independent MVN
        :rtype: gpytorch.distributions.MultivariateNormal
        """
        # Create batch distribution where all data are independent, but the tasks are dependent
        full_covar = self.lazy_covariance_matrix
        num_data, num_tasks = self.mean.shape[-2:]
        if self._interleaved:
            data_indices = torch.arange(0, num_data * num_tasks, num_tasks, device=full_covar.device).view(-1, 1, 1)
            task_indices = torch.arange(num_tasks, device=full_covar.device)
        else:
            data_indices = torch.arange(num_data, device=full_covar.device).view(-1, 1, 1)
            task_indices = torch.arange(0, num_data * num_tasks, num_data, device=full_covar.device)
        task_covars = full_covar[
            ..., data_indices + task_indices.unsqueeze(-2), data_indices + task_indices.unsqueeze(-1)
        ]
        return MultivariateNormal(self.mean, to_linear_operator(task_covars).add_jitter(jitter_val=jitter_val))

    @property
    def variance(self):
        var = super().variance
        if not self._interleaved:
            # flip shape of last two dimensions
            new_shape = self._output_shape[:-2] + self._output_shape[:-3:-1]
            return var.view(new_shape).transpose(-1, -2).contiguous()
        return var.view(self._output_shape)

    def __getitem__(self, idx) -> MultivariateNormal:
        """
        Constructs a new MultivariateNormal that represents a random variable
        modified by an indexing operation.

        The mean and covariance matrix arguments are indexed accordingly.

        :param Any idx: Index to apply to the mean. The covariance matrix is indexed accordingly.
        :returns: If indices specify a slice for samples and tasks, returns a
            MultitaskMultivariateNormal, else returns a MultivariateNormal.
        """

        # Normalize index to a tuple
        if not isinstance(idx, tuple):
            idx = (idx,)

        if ... in idx:
            # Replace ellipsis '...' with explicit indices
            ellipsis_location = idx.index(...)
            if ... in idx[ellipsis_location + 1 :]:
                raise IndexError("Only one ellipsis '...' is supported!")
            prefix = idx[:ellipsis_location]
            suffix = idx[ellipsis_location + 1 :]
            infix_length = self.mean.dim() - len(prefix) - len(suffix)
            if infix_length < 0:
                raise IndexError(f"Index {idx} has too many dimensions")
            idx = prefix + (slice(None),) * infix_length + suffix
        elif len(idx) == self.mean.dim() - 1:
            # Normalize indices ignoring the task-index to include it
            idx = idx + (slice(None),)

        new_mean = self.mean[idx]

        # We now create a covariance matrix appropriate for new_mean
        if len(idx) <= self.mean.dim() - 2:
            # We are only indexing the batch dimensions in this case
            return MultitaskMultivariateNormal(
                mean=new_mean,
                covariance_matrix=self.lazy_covariance_matrix[idx],
                interleaved=self._interleaved,
            )
        elif len(idx) > self.mean.dim():
            raise IndexError(f"Index {idx} has too many dimensions")
        else:
            # We have an index that extends over all dimensions
            batch_idx = idx[:-2]
            if self._interleaved:
                row_idx = idx[-2]
                col_idx = idx[-1]
                num_rows = self._output_shape[-2]
                num_cols = self._output_shape[-1]
            else:
                row_idx = idx[-1]
                col_idx = idx[-2]
                num_rows = self._output_shape[-1]
                num_cols = self._output_shape[-2]

            if isinstance(row_idx, int) and isinstance(col_idx, int):
                # Single sample with single task
                row_idx = _normalize_index(row_idx, num_rows)
                col_idx = _normalize_index(col_idx, num_cols)
                new_cov = DiagLinearOperator(
                    self.lazy_covariance_matrix.diagonal()[batch_idx + (row_idx * num_cols + col_idx,)]
                )
                return MultivariateNormal(mean=new_mean, covariance_matrix=new_cov)
            elif isinstance(row_idx, int) and isinstance(col_idx, slice):
                # A block of the covariance matrix
                row_idx = _normalize_index(row_idx, num_rows)
                col_idx = _normalize_slice(col_idx, num_cols)
                new_slice = slice(
                    col_idx.start + row_idx * num_cols,
                    col_idx.stop + row_idx * num_cols,
                    col_idx.step,
                )
                new_cov = self.lazy_covariance_matrix[batch_idx + (new_slice, new_slice)]
                return MultivariateNormal(mean=new_mean, covariance_matrix=new_cov)
            elif isinstance(row_idx, slice) and isinstance(col_idx, int):
                # A block of the reversely interleaved covariance matrix
                row_idx = _normalize_slice(row_idx, num_rows)
                col_idx = _normalize_index(col_idx, num_cols)
                new_slice = slice(row_idx.start + col_idx, row_idx.stop * num_cols + col_idx, row_idx.step * num_cols)
                new_cov = self.lazy_covariance_matrix[batch_idx + (new_slice, new_slice)]
                return MultivariateNormal(mean=new_mean, covariance_matrix=new_cov)
            elif (
                isinstance(row_idx, slice)
                and isinstance(col_idx, slice)
                and row_idx == col_idx == slice(None, None, None)
            ):
                new_cov = self.lazy_covariance_matrix[batch_idx]
                return MultitaskMultivariateNormal(
                    mean=new_mean,
                    covariance_matrix=new_cov,
                    interleaved=self._interleaved,
                    validate_args=False,
                )
            elif isinstance(row_idx, slice) or isinstance(col_idx, slice):
                # slice x slice or indices x slice or slice x indices
                if isinstance(row_idx, slice):
                    row_idx = torch.arange(num_rows)[row_idx]
                if isinstance(col_idx, slice):
                    col_idx = torch.arange(num_cols)[col_idx]
                row_grid, col_grid = torch.meshgrid(row_idx, col_idx, indexing="ij")
                indices = (row_grid * num_cols + col_grid).reshape(-1)
                new_cov = self.lazy_covariance_matrix[batch_idx + (indices,)][..., indices]
                return MultitaskMultivariateNormal(
                    mean=new_mean, covariance_matrix=new_cov, interleaved=self._interleaved, validate_args=False
                )
            else:
                # row_idx and col_idx have pairs of indices
                indices = row_idx * num_cols + col_idx
                new_cov = self.lazy_covariance_matrix[batch_idx + (indices,)][..., indices]
                return MultivariateNormal(
                    mean=new_mean,
                    covariance_matrix=new_cov,
                )

    def __repr__(self) -> str:
        return f"MultitaskMultivariateNormal(mean shape: {self._output_shape})"


def _normalize_index(i: int, dim_size: int) -> int:
    if i < 0:
        return dim_size + i
    else:
        return i


def _normalize_slice(s: slice, dim_size: int) -> slice:
    start = s.start
    if start is None:
        start = 0
    elif start < 0:
        start = dim_size + start
    stop = s.stop
    if stop is None:
        stop = dim_size
    elif stop < 0:
        stop = dim_size + stop
    step = s.step
    if step is None:
        step = 1
    return slice(start, stop, step)



#!/usr/bin/env python3

import pickle
import unittest

import torch

from gpytorch.kernels import HammingIMQKernel
from gpytorch.priors import GammaPrior


class TestHammingIMQKernel(unittest.TestCase):
    def create_seq(self, batch_size, seq_len, vocab_size):
        return torch.randint(0, vocab_size, (batch_size, seq_len))

    def create_seq_pairs(self, batch_size, seq_len, vocab_size):
        a = self.create_seq(batch_size, seq_len, vocab_size)
        b = self.create_seq(batch_size, seq_len, vocab_size)
        set_to_a = torch.rand(batch_size, seq_len) < 0.5
        b = torch.where(set_to_a, a, b)
        return a, b

    def test_computes_hamming_imq_function(self):
        """
        Create one-hot encoded discrete sequences and flatten them.
        Compute the pairwise Hamming distance $d$ between the sequences.
        Check the result of the kernel evaluation is
        $((1 + \alpha) / (\alpha + d))^{\beta}$.
        """
        vocab_size = 8
        seq_len = 4
        alpha = 2.0
        beta = 0.5
        kernel = HammingIMQKernel(vocab_size=vocab_size)
        kernel.initialize(alpha=alpha, beta=beta)
        kernel.eval()

        # Create two discrete sequences with some matches.
        a = torch.tensor([[7, 7, 7, 7], [5, 7, 3, 4]])
        b = torch.tensor(
            [
                [7, 5, 7, 4],
                [6, 7, 3, 7],
                [5, 7, 3, 4],
            ]
        )

        # Convert to one-hot representation.
        a_one_hot = torch.zeros(*a.shape, vocab_size)
        a_one_hot.scatter_(index=a.unsqueeze(-1), dim=-1, value=1)
        b_one_hot = torch.zeros(*b.shape, vocab_size)
        b_one_hot.scatter_(index=b.unsqueeze(-1), dim=-1, value=1)

        # Flatten the one-hot representations.
        a_one_hot_flat = a_one_hot.view(a.size(0), -1)
        b_one_hot_flat = b_one_hot.view(b.size(0), -1)

        # Compute the Hamming distance.
        d = seq_len - (a_one_hot.unsqueeze(-3) * b_one_hot.unsqueeze(-4)).sum(dim=(-1, -2))

        # Compute the kernel evaluation.
        actual = ((1 + alpha) / (alpha + d)) ** beta
        res = kernel(a_one_hot_flat, b_one_hot_flat).to_dense()

        # Check the result.
        self.assertLess(torch.norm(res - actual), 1e-5)

    def test_initialize_alpha(self):
        """
        Check that the kernel can be initialized with alpha.
        """
        alpha = 2.0
        kernel = HammingIMQKernel(vocab_size=8)
        kernel.initialize(alpha=alpha)
        actual_value = torch.tensor(alpha).view_as(kernel.alpha)
        self.assertLess(torch.norm(kernel.alpha - actual_value), 1e-5)

    def test_initialize_alpha_batch(self):
        batch_size = 2
        alpha = torch.rand(batch_size)
        kernel = HammingIMQKernel(vocab_size=8, batch_shape=torch.Size([batch_size]))
        kernel.initialize(alpha=alpha)
        actual_value = alpha.view_as(kernel.alpha)
        self.assertLess(torch.norm(kernel.alpha - actual_value), 1e-5)

    def test_initialize_beta(self):
        """
        Check that the kernel can be initialized with beta.
        """
        beta = 0.5
        kernel = HammingIMQKernel(vocab_size=8)
        kernel.initialize(beta=beta)
        actual_value = torch.tensor(beta).view_as(kernel.beta)
        self.assertLess(torch.norm(kernel.beta - actual_value), 1e-5)

    def test_initialize_beta_batch(self):
        batch_size = 2
        beta = torch.rand(batch_size)
        kernel = HammingIMQKernel(vocab_size=8, batch_shape=torch.Size([batch_size]))
        kernel.initialize(beta=beta)
        actual_value = beta.view_as(kernel.beta)
        self.assertLess(torch.norm(kernel.beta - actual_value), 1e-5)

    def create_kernel_with_prior(self, alpha_prior=None, beta_prior=None):
        return HammingIMQKernel(
            vocab_size=8,
            alpha_prior=alpha_prior,
            beta_prior=beta_prior,
        )

    def test_prior_type(self):
        self.create_kernel_with_prior()
        self.create_kernel_with_prior(
            alpha_prior=GammaPrior(1.0, 1.0),
            beta_prior=GammaPrior(1.0, 1.0),
        )
        self.assertRaises(TypeError, self.create_kernel_with_prior, 1)

    def test_pickle_with_prior(self):
        kernel = self.create_kernel_with_prior(
            alpha_prior=GammaPrior(1.0, 1.0),
            beta_prior=GammaPrior(1.0, 1.0),
        )
        pickle.loads(pickle.dumps(kernel))  # Should be able to pickle and unpickle with a prior.


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import unittest
from unittest.mock import MagicMock, patch

import linear_operator
import torch

import gpytorch
from gpytorch.kernels import InducingPointKernel, RBFKernel, ScaleKernel


class _TestModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y):
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ZeroMean()
        self.covar_module = InducingPointKernel(
            ScaleKernel(RBFKernel(ard_num_dims=3)),
            inducing_points=torch.randn(512, 3),
            likelihood=likelihood,
        )

    def forward(self, input):
        mean = self.mean_module(input)
        covar = self.covar_module(input)
        return gpytorch.distributions.MultivariateNormal(mean, covar)


class TestInducingPointKernel(unittest.TestCase):
    def test_kernel_output(self):
        train_x = torch.randn(1000, 3)
        train_y = torch.randn(1000)
        test_x = torch.randn(500, 3)
        model = _TestModel(train_x, train_y)

        # Make sure that the prior kernel is the correct type
        model.train()
        output = model(train_x).lazy_covariance_matrix.evaluate_kernel()
        self.assertIsInstance(output, linear_operator.operators.LowRankRootLinearOperator)

        # Make sure that the prior predictive kernel is the correct type
        model.train()
        output = model.likelihood(model(train_x)).lazy_covariance_matrix.evaluate_kernel()
        self.assertIsInstance(output, linear_operator.operators.LowRankRootAddedDiagLinearOperator)

        # Make sure we're calling the correct prediction strategy
        _wrapped_ps = MagicMock(wraps=gpytorch.models.exact_prediction_strategies.SGPRPredictionStrategy)
        with patch("gpytorch.models.exact_prediction_strategies.SGPRPredictionStrategy", new=_wrapped_ps) as ps_mock:
            model.eval()
            output = model.likelihood(model(test_x))
            _ = output.mean + output.variance  # Compute something to break through any lazy evaluations
            self.assertTrue(ps_mock.called)

        # Check whether changing diagonal correction makes a difference (ensuring that cache is cleared)
        model.train()
        model.eval()
        with gpytorch.settings.sgpr_diagonal_correction(True), torch.no_grad():
            output_mean_correct = model(test_x).mean
        model.train()
        model.eval()
        with gpytorch.settings.sgpr_diagonal_correction(False), torch.no_grad():
            output_mean_no_correct = model(test_x).mean
        self.assertNotAlmostEqual(output_mean_correct.sum().item(), output_mean_no_correct.sum().item())



#!/usr/bin/env python3

import io
import os
import re
import sys

from setuptools import find_packages, setup

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 10

# Check for python version
if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    error = (
        "Your version of python ({major}.{minor}) is too old. You need python >= {required_major}.{required_minor}."
    ).format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
        required_minor=REQUIRED_MINOR,
        required_major=REQUIRED_MAJOR,
    )
    sys.exit(error)


# Get version
def find_version(*file_paths):
    try:
        with io.open(os.path.join(os.path.dirname(__file__), *file_paths), encoding="utf8") as fp:
            version_file = fp.read()
        version_match = re.search(r"^__version__ = version = ['\"]([^'\"]*)['\"]", version_file, re.M)
        return version_match.group(1)
    except Exception:
        return None


readme = open("README.md").read()


torch_min = "2.0"
install_requires = [
    "jaxtyping",
    "mpmath>=0.19,<=1.3",  # avoid incompatibiltiy with torch+sympy with mpmath 1.4
    "scikit-learn",
    "scipy>=1.6.0",
    "linear_operator>=0.6",
]
# if recent dev version of PyTorch is installed, no need to install stable
try:
    import torch

    if torch.__version__ >= torch_min:
        install_requires = [">=".join(["torch", torch_min])] + install_requires
except ImportError:
    pass


# Run the setup
setup(
    name="gpytorch",
    version=find_version("gpytorch", "version.py"),
    description="An implementation of Gaussian Processes in Pytorch",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Jake Gardner, Geoff Pleiss",
    url="https://gpytorch.ai",
    author_email="gpleiss@gmail.com",
    project_urls={
        "Documentation": "https://gpytorch.readthedocs.io",
        "Source": "https://github.com/cornellius-gp/gpytorch/",
    },
    license="MIT",
    classifiers=["Development Status :: 5 - Production/Stable", "Programming Language :: Python :: 3"],
    packages=find_packages(exclude=["test", "test.*"]),
    python_requires=f">={REQUIRED_MAJOR}.{REQUIRED_MINOR}",
    install_requires=install_requires,
    extras_require={
        "dev": ["pre-commit", "setuptools_scm", "twine", "ufmt"],
        "docs": [
            "ipykernel<=6.17.1",
            "ipython<=8.6.0",
            "m2r2<=0.3.3.post2",
            "nbclient<=0.7.3",
            "nbformat<=5.8.0",
            "nbsphinx<=0.9.1",
            "lxml_html_clean",
            "platformdirs<=3.2.0",
            "setuptools_scm<=7.1.0",
            "sphinx<=6.2.1",
            "sphinx_autodoc_typehints<=1.23.0",
            "sphinx_rtd_theme<0.5",
        ],
        "examples": ["ipython", "jupyter", "matplotlib", "scipy", "torchvision", "tqdm"],
        "keops": ["pykeops>=1.1.1"],
        "pyro": ["pyro-ppl>=1.8"],
        "test": ["flake8==4.0.1", "flake8-print==4.0.0", "pytest", "nbval"],
    },
    test_suite="test",
)



#!/usr/bin/env python3

import math

from linear_operator.operators import KernelLinearOperator
from torch import Tensor

from ..periodic_kernel import PeriodicKernel as GPeriodicKernel
from .keops_kernel import _Anysor, _lazify_and_expand_inputs, KeOpsKernel


def _covar_func(x1: _Anysor, x2: _Anysor, lengthscale: Tensor, **kwargs) -> _Anysor:
    x1_, x2_ = _lazify_and_expand_inputs(x1, x2)
    lengthscale = lengthscale[..., None, None, 0, :]  # 1 x 1 x ndim
    # do not use .power(2.0) as it gives NaN values on cuda
    # seems related to https://github.com/getkeops/keops/issues/112
    K = ((((x1_ - x2_).abs().sin()) ** 2) * (-2.0 / lengthscale)).sum(-1).exp()
    return K


# subclass from original periodic kernel to reduce code duplication
class PeriodicKernel(KeOpsKernel, GPeriodicKernel):
    """
    Implements the Periodic Kernel using KeOps as a driver for kernel matrix multiplies.

    This class can be used as a drop in replacement for :class:`gpytorch.kernels.PeriodicKernel` in most cases,
    and supports the same arguments.

    :param ard_num_dims: (Default: `None`) Set this if you want a separate lengthscale for each
        input dimension. It should be `d` if x1 is a `... x n x d` matrix.
    :type ard_num_dims: int, optional
    :param batch_shape: (Default: `None`) Set this if you want a separate lengthscale for each
         batch of input data. It should be `torch.Size([b1, b2])` for a `b1 x b2 x n x m` kernel output.
    :type batch_shape: torch.Size, optional
    :param active_dims: (Default: `None`) Set this if you want to
        compute the covariance of only a few input dimensions. The ints
        corresponds to the indices of the dimensions.
    :type active_dims: Tuple(int)
    :param period_length_prior: (Default: `None`)
        Set this if you want to apply a prior to the period length parameter.
    :type period_length_prior: ~gpytorch.priors.Prior, optional
    :param period_length_constraint: (Default: `Positive`) Set this if you want
        to apply a constraint to the period length parameter.
    :type period_length_constraint: ~gpytorch.constraints.Interval, optional
    :param lengthscale_prior: (Default: `None`)
        Set this if you want to apply a prior to the lengthscale parameter.
    :type lengthscale_prior: ~gpytorch.priors.Prior, optional
    :param lengthscale_constraint: (Default: `Positive`) Set this if you want
        to apply a constraint to the lengthscale parameter.
    :type lengthscale_constraint: ~gpytorch.constraints.Interval, optional
    :param eps: (Default: 1e-6) The minimum value that the lengthscale can take (prevents divide by zero errors).
    :type eps: float, optional

    :var torch.Tensor period_length: The period length parameter. Size/shape of parameter depends on the
        ard_num_dims and batch_shape arguments.
    """

    has_lengthscale = True

    def forward(self, x1: Tensor, x2: Tensor, diag: bool = False, **kwargs) -> KernelLinearOperator:
        x1_ = x1.div(self.period_length / math.pi)
        x2_ = x2.div(self.period_length / math.pi)
        # return KernelLinearOperator inst only when calculating the whole covariance matrix
        # pass any parameters which are used inside _covar_func as *args to get gradients computed for them
        res = KernelLinearOperator(x1_, x2_, lengthscale=self.lengthscale, covar_func=_covar_func, **kwargs)

        # TODO: diag=True mode will be removed with the GpyTorch 2.0 PR to remove LazyEvaluatedKernelTensor
        # (it will be replaced by a `_symmetric_diag` method for quickly computing the diagonals of symmetric matrices)
        if diag:
            return res.diagonal(dim1=-1, dim2=-2)
        else:
            return res



#!/usr/bin/env python3

import unittest

import torch
from linear_operator.operators import DiagLinearOperator

from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import InducingPointKernelAddedLossTerm
from gpytorch.test.base_test_case import BaseTestCase


class TestInducingPointKernelAddedLossTerm(BaseTestCase, unittest.TestCase):
    def test_added_loss_term(self):
        # This loss term won't usually be called with diagonal MVNs
        # However, the loss term only accesses the diagonals of the MVN covariance matrices
        # So we're simplifying the setup for the unit test
        prior_dist = MultivariateNormal(torch.zeros(5), DiagLinearOperator(torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0])))
        variational_dist = MultivariateNormal(
            torch.zeros(5), DiagLinearOperator(torch.tensor([0.6, 0.7, 0.8, 0.9, 1.0]))
        )
        likelihood = GaussianLikelihood()
        likelihood.noise = 0.01

        added_loss_term = InducingPointKernelAddedLossTerm(prior_dist, variational_dist, likelihood)
        self.assertAllClose(added_loss_term.loss(), torch.tensor(-50.0))

    def test_added_loss_term_batch(self):
        prior_dist = MultivariateNormal(
            torch.zeros(2, 5), DiagLinearOperator(torch.tensor([[1.0, 1.0, 1.0, 1.0, 1.0], [1.0, 1.0, 1.0, 1.0, 1.0]]))
        )
        variational_dist = MultivariateNormal(
            torch.zeros(2, 5),
            DiagLinearOperator(torch.tensor([[0.6, 0.7, 0.8, 0.9, 1.0], [0.8, 0.85, 0.9, 0.95, 1.0]])),
        )
        likelihood = GaussianLikelihood(batch_shape=torch.Size([3, 1]))
        likelihood.noise = torch.Tensor([[0.01], [0.1], [1.0]])

        added_loss_term = InducingPointKernelAddedLossTerm(prior_dist, variational_dist, likelihood)
        self.assertAllClose(added_loss_term.loss(), torch.tensor([[-50.0, -25.0], [-5.0, -2.5], [-0.5, -0.25]]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from abc import abstractproperty
from unittest.mock import MagicMock, patch

import linear_operator
import torch

import gpytorch

from .base_test_case import BaseTestCase


class VariationalTestCase(BaseTestCase):
    def _make_model_and_likelihood(
        self,
        num_inducing=16,
        batch_shape=torch.Size([]),
        inducing_batch_shape=torch.Size([]),
        strategy_cls=gpytorch.variational.VariationalStrategy,
        distribution_cls=gpytorch.variational.CholeskyVariationalDistribution,
        constant_mean=True,
    ):
        class _SVGPRegressionModel(gpytorch.models.ApproximateGP):
            def __init__(self, inducing_points):
                variational_distribution = distribution_cls(num_inducing, batch_shape=batch_shape)
                variational_strategy = strategy_cls(
                    self,
                    inducing_points,
                    variational_distribution,
                    learn_inducing_locations=True,
                )
                super().__init__(variational_strategy)
                if constant_mean:
                    self.mean_module = gpytorch.means.ConstantMean()
                    self.mean_module.initialize(constant=1.0)
                else:
                    self.mean_module = gpytorch.means.ZeroMean()
                self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

            def forward(self, x):
                mean_x = self.mean_module(x)
                covar_x = self.covar_module(x)
                latent_pred = gpytorch.distributions.MultivariateNormal(mean_x, covar_x)
                return latent_pred

        inducing_points = torch.randn(num_inducing, 2).repeat(*inducing_batch_shape, 1, 1)
        return _SVGPRegressionModel(inducing_points), self.likelihood_cls()

    def _training_iter(
        self,
        model,
        likelihood,
        batch_shape=torch.Size([]),
        mll_cls=gpytorch.mlls.VariationalELBO,
        cuda=False,
    ):
        train_x = torch.randn(*batch_shape, 32, 2).clamp(-2.5, 2.5)
        train_y = torch.linspace(-1, 1, self.event_shape[0])
        train_y = train_y.view(self.event_shape[0], *([1] * (len(self.event_shape) - 1)))
        train_y = train_y.expand(*self.event_shape)
        mll = mll_cls(likelihood, model, num_data=train_x.size(-2))
        if cuda:
            train_x = train_x.cuda()
            train_y = train_y.cuda()
            model = model.cuda()
            likelihood = likelihood.cuda()

        # Single optimization iteration
        model.train()
        likelihood.train()
        output = model(train_x)
        loss = -mll(output, train_y)
        loss.sum().backward()

        # Make sure we have gradients for all parameters
        for _, param in model.named_parameters():
            self.assertTrue(param.grad is not None)
            self.assertGreater(param.grad.norm().item(), 0)
        for _, param in likelihood.named_parameters():
            self.assertTrue(param.grad is not None)
            self.assertGreater(param.grad.norm().item(), 0)

        return output, loss

    def _eval_iter(self, model, batch_shape=torch.Size([]), cuda=False):
        test_x = torch.randn(*batch_shape, 32, 2).clamp(-2.5, 2.5)
        if cuda:
            test_x = test_x.cuda()
            model = model.cuda()

        # Single optimization iteration
        model.eval()
        with torch.no_grad():
            output = model(test_x)

        return output

    def _fantasy_iter(
        self,
        model,
        likelihood,
        batch_shape=torch.Size([]),
        cuda=False,
        num_fant=10,
        covar_module=None,
        mean_module=None,
    ):
        model.likelihood = likelihood
        val_x = torch.randn(*batch_shape, num_fant, 2).clamp(-2.5, 2.5)
        val_y = torch.linspace(-1, 1, num_fant)
        val_y = val_y.view(num_fant, *([1] * (len(self.event_shape) - 1)))
        val_y = val_y.expand(*batch_shape, num_fant, *self.event_shape[1:])
        if cuda:
            model = model.cuda()
            val_x = val_x.cuda()
            val_y = val_y.cuda()
        updated_model = model.get_fantasy_model(val_x, val_y, covar_module=covar_module, mean_module=mean_module)
        return updated_model

    @abstractproperty
    def batch_shape(self):
        raise NotImplementedError

    @abstractproperty
    def distribution_cls(self):
        raise NotImplementedError

    @property
    def event_shape(self):
        return torch.Size([32])

    @property
    def likelihood_cls(self):
        return gpytorch.likelihoods.GaussianLikelihood

    @abstractproperty
    def mll_cls(self):
        raise NotImplementedError

    @abstractproperty
    def strategy_cls(self):
        raise NotImplementedError

    @property
    def cuda(self):
        return False

    def test_eval_iteration(
        self,
        data_batch_shape=None,
        inducing_batch_shape=None,
        model_batch_shape=None,
        eval_data_batch_shape=None,
        expected_batch_shape=None,
    ):
        # Batch shapes
        model_batch_shape = model_batch_shape if model_batch_shape is not None else self.batch_shape
        data_batch_shape = data_batch_shape if data_batch_shape is not None else self.batch_shape
        inducing_batch_shape = inducing_batch_shape if inducing_batch_shape is not None else self.batch_shape
        expected_batch_shape = expected_batch_shape if expected_batch_shape is not None else self.batch_shape
        eval_data_batch_shape = eval_data_batch_shape if eval_data_batch_shape is not None else self.batch_shape

        # Mocks
        _wrapped_cholesky = MagicMock(wraps=torch.linalg.cholesky_ex)
        _wrapped_cg = MagicMock(wraps=linear_operator.utils.linear_cg)
        _wrapped_ciq = MagicMock(wraps=linear_operator.utils.contour_integral_quad)
        _cholesky_mock = patch("torch.linalg.cholesky_ex", new=_wrapped_cholesky)
        _cg_mock = patch("linear_operator.utils.linear_cg", new=_wrapped_cg)
        _ciq_mock = patch("linear_operator.utils.contour_integral_quad", new=_wrapped_ciq)

        # Make model and likelihood
        model, likelihood = self._make_model_and_likelihood(
            batch_shape=model_batch_shape,
            inducing_batch_shape=inducing_batch_shape,
            distribution_cls=self.distribution_cls,
            strategy_cls=self.strategy_cls,
        )

        # Do one forward pass
        self._training_iter(model, likelihood, data_batch_shape, mll_cls=self.mll_cls, cuda=self.cuda)

        # Now do evaluation
        with _cholesky_mock as cholesky_mock, _cg_mock as cg_mock, _ciq_mock as ciq_mock:
            # Iter 1
            _ = self._eval_iter(model, eval_data_batch_shape, cuda=self.cuda)
            output = self._eval_iter(model, eval_data_batch_shape, cuda=self.cuda)
            self.assertEqual(output.batch_shape, expected_batch_shape)
            self.assertEqual(output.event_shape, self.event_shape)
            return cg_mock, cholesky_mock, ciq_mock

    def test_eval_smaller_pred_batch(self):
        return self.test_eval_iteration(
            model_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            inducing_batch_shape=(torch.Size([3, 1]) + self.batch_shape),
            data_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            eval_data_batch_shape=(torch.Size([4]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
        )

    def test_eval_larger_pred_batch(self):
        return self.test_eval_iteration(
            model_batch_shape=(torch.Size([4]) + self.batch_shape),
            inducing_batch_shape=(self.batch_shape),
            data_batch_shape=(torch.Size([4]) + self.batch_shape),
            eval_data_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
        )

    def test_training_iteration(
        self,
        data_batch_shape=None,
        inducing_batch_shape=None,
        model_batch_shape=None,
        expected_batch_shape=None,
        constant_mean=True,
    ):
        # Batch shapes
        model_batch_shape = model_batch_shape if model_batch_shape is not None else self.batch_shape
        data_batch_shape = data_batch_shape if data_batch_shape is not None else self.batch_shape
        inducing_batch_shape = inducing_batch_shape if inducing_batch_shape is not None else self.batch_shape
        expected_batch_shape = expected_batch_shape if expected_batch_shape is not None else self.batch_shape

        # Mocks
        _wrapped_cholesky = MagicMock(wraps=torch.linalg.cholesky_ex)
        _wrapped_cg = MagicMock(wraps=linear_operator.utils.linear_cg)
        _wrapped_ciq = MagicMock(wraps=linear_operator.utils.contour_integral_quad)
        _cholesky_mock = patch("torch.linalg.cholesky_ex", new=_wrapped_cholesky)
        _cg_mock = patch("linear_operator.utils.linear_cg", new=_wrapped_cg)
        _ciq_mock = patch("linear_operator.utils.contour_integral_quad", new=_wrapped_ciq)

        # Make model and likelihood
        model, likelihood = self._make_model_and_likelihood(
            batch_shape=model_batch_shape,
            inducing_batch_shape=inducing_batch_shape,
            distribution_cls=self.distribution_cls,
            strategy_cls=self.strategy_cls,
            constant_mean=constant_mean,
        )

        # Do forward pass
        with _cholesky_mock as cholesky_mock, _cg_mock as cg_mock, _ciq_mock as ciq_mock:
            # Iter 1
            self.assertEqual(model.variational_strategy.variational_params_initialized.item(), 0)
            self._training_iter(
                model,
                likelihood,
                data_batch_shape,
                mll_cls=self.mll_cls,
                cuda=self.cuda,
            )
            self.assertEqual(model.variational_strategy.variational_params_initialized.item(), 1)
            # Iter 2
            output, loss = self._training_iter(
                model,
                likelihood,
                data_batch_shape,
                mll_cls=self.mll_cls,
                cuda=self.cuda,
            )
            self.assertEqual(output.batch_shape, expected_batch_shape)
            self.assertEqual(output.event_shape, self.event_shape)
            self.assertEqual(loss.shape, expected_batch_shape)
            return cg_mock, cholesky_mock, ciq_mock

    def test_training_iteration_batch_inducing(self):
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            data_batch_shape=self.batch_shape,
            inducing_batch_shape=(torch.Size([3]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_training_iteration_batch_data(self):
        return self.test_training_iteration(
            model_batch_shape=self.batch_shape,
            inducing_batch_shape=self.batch_shape,
            data_batch_shape=(torch.Size([3]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_training_iteration_batch_model(self):
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            inducing_batch_shape=self.batch_shape,
            data_batch_shape=self.batch_shape,
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_training_all_batch_zero_mean(self):
        return self.test_training_iteration(
            model_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            inducing_batch_shape=(torch.Size([3, 1]) + self.batch_shape),
            data_batch_shape=(torch.Size([4]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3, 4]) + self.batch_shape),
            constant_mean=False,
        )

    def test_fantasy_call(
        self,
        data_batch_shape=None,
        inducing_batch_shape=None,
        model_batch_shape=None,
        expected_batch_shape=None,
        constant_mean=True,
    ):
        # Batch shapes
        model_batch_shape = model_batch_shape if model_batch_shape is not None else self.batch_shape
        data_batch_shape = data_batch_shape if data_batch_shape is not None else self.batch_shape
        inducing_batch_shape = inducing_batch_shape if inducing_batch_shape is not None else self.batch_shape
        expected_batch_shape = expected_batch_shape if expected_batch_shape is not None else self.batch_shape

        num_inducing = 16
        num_fant = 10
        # Make model and likelihood
        model, likelihood = self._make_model_and_likelihood(
            batch_shape=model_batch_shape,
            inducing_batch_shape=inducing_batch_shape,
            distribution_cls=self.distribution_cls,
            strategy_cls=self.strategy_cls,
            constant_mean=constant_mean,
            num_inducing=num_inducing,
        )

        # we iterate through the covar and mean module possible settings
        covar_mean_options = [
            {"covar_module": None, "mean_module": None},
            {"covar_module": gpytorch.kernels.MaternKernel(), "mean_module": gpytorch.means.ZeroMean()},
        ]
        for cm_dict in covar_mean_options:
            fant_model = self._fantasy_iter(
                model, likelihood, data_batch_shape, self.cuda, num_fant=num_fant, **cm_dict
            )
            self.assertTrue(isinstance(fant_model, gpytorch.models.ExactGP))

            # we check to ensure setting the covar_module and mean_modules are okay
            if cm_dict["covar_module"] is None:
                self.assertEqual(type(fant_model.covar_module), type(model.covar_module))
            else:
                self.assertNotEqual(type(fant_model.covar_module), type(model.covar_module))
            if cm_dict["mean_module"] is None:
                self.assertEqual(type(fant_model.mean_module), type(model.mean_module))
            else:
                self.assertNotEqual(type(fant_model.mean_module), type(model.mean_module))

            # now we check to ensure the shapes of the fantasy strategy are correct
            self.assertTrue(fant_model.prediction_strategy is not None)
            for key in fant_model.prediction_strategy._memoize_cache.keys():
                if key[0] == "mean_cache":
                    break
            mean_cache = fant_model.prediction_strategy._memoize_cache[key]
            self.assertEqual(mean_cache.shape, torch.Size([*expected_batch_shape, num_inducing + num_fant]))

        # we remove the mean_module and covar_module and check for errors
        del model.mean_module
        with self.assertRaises(ModuleNotFoundError):
            self._fantasy_iter(model, likelihood, data_batch_shape, self.cuda, num_fant=num_fant)

        model.mean_module = gpytorch.means.ZeroMean()
        del model.covar_module
        with self.assertRaises(ModuleNotFoundError):
            self._fantasy_iter(model, likelihood, data_batch_shape, self.cuda, num_fant=num_fant)

        # finally we check to ensure failure for a non-gaussian likelihood
        with self.assertRaises(NotImplementedError):
            self._fantasy_iter(
                model,
                gpytorch.likelihoods.BernoulliLikelihood(),
                data_batch_shape,
                self.cuda,
                num_fant=num_fant,
            )

    def test_fantasy_call_batch_inducing(self):
        return self.test_fantasy_call(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            data_batch_shape=self.batch_shape,
            inducing_batch_shape=(torch.Size([3]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_fantasy_call_batch_data(self):
        return self.test_fantasy_call(
            model_batch_shape=self.batch_shape,
            inducing_batch_shape=self.batch_shape,
            data_batch_shape=(torch.Size([3]) + self.batch_shape),
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )

    def test_fantasy_call_batch_model(self):
        return self.test_fantasy_call(
            model_batch_shape=(torch.Size([3]) + self.batch_shape),
            inducing_batch_shape=self.batch_shape,
            data_batch_shape=self.batch_shape,
            expected_batch_shape=(torch.Size([3]) + self.batch_shape),
        )



#!/usr/bin/env python3

from typing import Any, Optional

import torch

from ..priors import Prior
from .mean import Mean


class ConstantMeanGradGrad(Mean):
    r"""
    A (non-zero) constant prior mean function and its first and second derivatives, i.e.:

    .. math::

        \mu(\mathbf x) &= C \\
        \nabla \mu(\mathbf x) &= \mathbf 0 \\
        \nabla^2 \mu(\mathbf x) &= \mathbf 0

    where :math:`C` is a learned constant.

    :param prior: Prior for constant parameter :math:`C`.
    :type prior: ~gpytorch.priors.Prior, optional
    :param batch_shape: The batch shape of the learned constant(s) (default: []).
    :type batch_shape: torch.Size, optional

    :var torch.Tensor constant: :math:`C` parameter
    """

    def __init__(
        self,
        prior: Optional[Prior] = None,
        batch_shape: torch.Size = torch.Size(),
        **kwargs: Any,
    ):
        super(ConstantMeanGradGrad, self).__init__()
        self.batch_shape = batch_shape
        self.register_parameter(name="constant", parameter=torch.nn.Parameter(torch.zeros(*batch_shape, 1)))
        if prior is not None:
            self.register_prior("mean_prior", prior, "constant")

    def forward(self, input):
        batch_shape = torch.broadcast_shapes(self.batch_shape, input.shape[:-2])
        mean = self.constant.unsqueeze(-1).expand(*batch_shape, input.size(-2), 2 * input.size(-1) + 1).contiguous()
        mean[..., 1:] = 0
        return mean



#!/usr/bin/env python3

from abc import ABC
from typing import Any, Mapping

from torch.distributions import TransformedDistribution
from torch.nn import Module

from ..distributions import Distribution
from .utils import _load_transformed_to_base_dist


TRANSFORMED_ERROR_MSG = """Priors of TransformedDistributions should not have their \
'_transformed' attributes modified, these are just copies of the base attribute. \
Please modify the base attribute (e.g. {}) instead."""


class Prior(Distribution, Module, ABC):
    """
    Base class for Priors in GPyTorch.
    In GPyTorch, a parameter can be assigned a prior by passing it as the `prior` argument to
    :func:`~gpytorch.module.register_parameter`. GPyTorch performs internal bookkeeping of priors,
    and for each parameter with a registered prior includes the log probability of the parameter under its
    respective prior in computing the Marginal Log-Likelihood.
    """

    def transform(self, x):
        return self._transform(x) if self._transform is not None else x

    def log_prob(self, x):
        r"""
        :return: log-probability of the parameter value under the prior
        :rtype: torch.Tensor
        """
        return super(Prior, self).log_prob(self.transform(x))

    def load_state_dict(self, state_dict: Mapping[str, Any], *args, **kwargs):
        Module.load_state_dict(self, state_dict, *args, **kwargs)
        if isinstance(self, TransformedDistribution):
            _load_transformed_to_base_dist(self)

    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(self, name) and "_transformed_" in name:
            base_attr_name = name.replace("_transformed_", "")
            raise AttributeError(TRANSFORMED_ERROR_MSG.format(base_attr_name))

        elif hasattr(self, f"_transformed_{name}"):
            self.base_dist.__setattr__(name, value)
            super().__setattr__(f"_transformed_{name}", value)

        else:
            return super().__setattr__(name, value)



#!/usr/bin/env python3

import os
import random
import unittest

import torch
from torch import optim

import gpytorch
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import AdditiveStructureKernel, GridInterpolationKernel, RBFKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ZeroMean

n = 20
train_x = torch.zeros(pow(n, 2), 2)
for i in range(n):
    for j in range(n):
        train_x[i * n + j][0] = float(i) / (n - 1)
        train_x[i * n + j][1] = float(j) / (n - 1)
train_y = torch.sin(train_x[:, 0]) + torch.cos(train_x[:, 1])
train_y = train_y + torch.randn_like(train_y).div_(20.0)

m = 10
test_x = torch.zeros(pow(m, 2), 2)
for i in range(m):
    for j in range(m):
        test_x[i * m + j][0] = float(i) / (m - 1)
        test_x[i * m + j][1] = float(j) / (m - 1)
test_y = torch.sin(test_x[:, 0]) + torch.cos(test_x[:, 1])


# All tests that pass with the exact kernel should pass with the interpolated kernel.
class GPRegressionModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(GPRegressionModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = ZeroMean()
        self.base_covar_module = ScaleKernel(RBFKernel(ard_num_dims=2))
        self.covar_module = AdditiveStructureKernel(
            GridInterpolationKernel(self.base_covar_module, grid_size=100, num_dims=1), num_dims=2
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class TestKISSGPAdditiveRegression(unittest.TestCase):
    def setUp(self):
        if os.getenv("UNLOCK_SEED") is None or os.getenv("UNLOCK_SEED").lower() == "false":
            self.rng_state = torch.get_rng_state()
            torch.manual_seed(1)
            if torch.cuda.is_available():
                torch.cuda.manual_seed_all(1)
            random.seed(1)

    def tearDown(self):
        if hasattr(self, "rng_state"):
            torch.set_rng_state(self.rng_state)

    def test_kissgp_gp_mean_abs_error(self, toeplitz=False):
        likelihood = GaussianLikelihood()
        gp_model = GPRegressionModel(train_x, train_y, likelihood)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, gp_model)

        with gpytorch.settings.max_preconditioner_size(10), gpytorch.settings.max_cg_iterations(50):
            with gpytorch.settings.fast_pred_var(), gpytorch.settings.use_toeplitz(toeplitz):
                # Optimize the model
                gp_model.train()
                likelihood.train()

                optimizer = optim.Adam(gp_model.parameters(), lr=0.01)
                optimizer.n_iter = 0
                for _ in range(15):
                    optimizer.zero_grad()
                    output = gp_model(train_x)
                    loss = -mll(output, train_y)
                    loss.backward()
                    optimizer.n_iter += 1
                    optimizer.step()

                    for param in gp_model.parameters():
                        self.assertTrue(param.grad is not None)
                        self.assertGreater(param.grad.norm().item(), 0)

                # Test the model
                gp_model.eval()
                likelihood.eval()

                test_preds = likelihood(gp_model(test_x)).mean
                mean_abs_error = torch.mean(torch.abs(test_y - test_preds))
                self.assertLess(mean_abs_error.squeeze().item(), 0.2)


if __name__ == "__main__":
    unittest.main()



from math import pi
from typing import Optional

import torch

from ..distributions import MultitaskMultivariateNormal, MultivariateNormal

pi = torch.tensor(pi)


def mean_absolute_error(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
):
    """
    Mean absolute error.
    """
    combine_dim = -2 if isinstance(pred_dist, MultitaskMultivariateNormal) else -1
    return torch.abs(pred_dist.mean - test_y).mean(dim=combine_dim)


def mean_squared_error(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
    squared: bool = True,
):
    """
    Mean squared error.
    """
    combine_dim = -2 if isinstance(pred_dist, MultitaskMultivariateNormal) else -1
    res = torch.square(pred_dist.mean - test_y).mean(dim=combine_dim)
    if not squared:
        return res**0.5
    return res


def standardized_mean_squared_error(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
):
    """Standardized mean squared error.

    Standardizes the mean squared error by the variance of the test data.
    """
    return mean_squared_error(pred_dist, test_y, squared=True) / test_y.var()


def negative_log_predictive_density(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
):
    """Negative log predictive density.

    Computes the negative predictive log density normalized by the size of the test data.
    """
    combine_dim = -2 if isinstance(pred_dist, MultitaskMultivariateNormal) else -1
    return -pred_dist.log_prob(test_y) / test_y.shape[combine_dim]


def mean_standardized_log_loss(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
    train_y: Optional[torch.Tensor] = None,
):
    """
    Mean standardized log loss.

    Computes the average *standardized* log loss, which subtracts the loss obtained
    under the trivial model which predicts with the mean and variance of the training
    data from the mean log loss. See p.23 of Rasmussen and Williams (2006).

    If no training data is supplied, the mean log loss is computed.
    """
    combine_dim = -2 if isinstance(pred_dist, MultitaskMultivariateNormal) else -1

    f_mean = pred_dist.mean
    f_var = pred_dist.variance
    loss_model = (0.5 * torch.log(2 * pi * f_var) + torch.square(test_y - f_mean) / (2 * f_var)).mean(dim=combine_dim)
    res = loss_model

    if train_y is not None:
        data_mean = train_y.mean(dim=combine_dim)
        data_var = train_y.var()
        loss_trivial_model = (
            0.5 * torch.log(2 * pi * data_var) + torch.square(test_y - data_mean) / (2 * data_var)
        ).mean(dim=combine_dim)
        res = res - loss_trivial_model

    return res


def quantile_coverage_error(
    pred_dist: MultivariateNormal,
    test_y: torch.Tensor,
    quantile: float = 95.0,
):
    """
    Quantile coverage error.
    """
    if quantile <= 0 or quantile >= 100:
        raise NotImplementedError("Quantile must be between 0 and 100")
    combine_dim = -2 if isinstance(pred_dist, MultitaskMultivariateNormal) else -1
    standard_normal = torch.distributions.Normal(loc=0.0, scale=1.0)
    deviation = standard_normal.icdf(torch.as_tensor(0.5 + 0.5 * (quantile / 100)))
    lower = pred_dist.mean - deviation * pred_dist.stddev
    upper = pred_dist.mean + deviation * pred_dist.stddev
    n_samples_within_bounds = ((test_y > lower) * (test_y < upper)).sum(combine_dim)
    fraction = n_samples_within_bounds / test_y.shape[combine_dim]
    return torch.abs(fraction - quantile / 100)
